from __future__ import annotations
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Any, ClassVar, Generic
from .qlobjs import REPO_FIELDS, Object
from .types import ActivityType, Affiliation, CursorDict, RepoActivity, Repository
from .util import BogusEventError, NotFoundError, T, log

PAGE_SIZE = 50


@dataclass
class QueryManager(Generic[T]):
    has_next_page: bool = field(init=False, default=True)

    @abstractmethod
    def make_query(self) -> tuple[str, dict[str, Any]]:
        ...

    @abstractmethod
    def parse_response(self, data: Any) -> list[T]:
        ...


@dataclass
class ReposQuery(QueryManager[Repository]):
    PATH: ClassVar[tuple[str, ...]]
    cursor: str | None = field(init=False, default=None)

    def parse_response(self, data: Any) -> list[Repository]:
        root = data["data"]
        for p in self.PATH:
            root = root[p]
        new_cursor = root["pageInfo"]["endCursor"]
        if new_cursor is not None:
            assert isinstance(new_cursor, str)
            self.cursor = new_cursor
        self.has_next_page = root["pageInfo"]["hasNextPage"]
        return [Repository.from_node(node) for node in root["nodes"]]


@dataclass
class ViewersReposQuery(ReposQuery):
    PATH: ClassVar[tuple[str, ...]] = ("viewer", "repositories")
    affiliations: list[Affiliation]

    def make_query(self) -> tuple[str, dict[str, Any]]:
        q = str(
            Object(
                "query",
                {
                    "$page_size": "Int!",
                    "$affiliations": "[RepositoryAffiliation!]",
                    "$cursor": "String",
                },
                Object(
                    "viewer",
                    {},
                    Object(
                        "repositories",
                        {
                            "ownerAffiliations": "$affiliations",
                            "orderBy": "{field: NAME, direction: ASC}",
                            "first": "$page_size",
                            "after": "$cursor",
                        },
                        Object("nodes", {}, *REPO_FIELDS),
                        Object("pageInfo", {}, "endCursor", "hasNextPage"),
                    ),
                ),
            )
        )
        variables = {
            "affiliations": [aff.value for aff in self.affiliations],
            "page_size": PAGE_SIZE,
            "cursor": self.cursor,
        }
        return (q, variables)


@dataclass
class OwnersReposQuery(ReposQuery):
    PATH: ClassVar[tuple[str, ...]] = ("repositoryOwner", "repositories")
    owner: str

    def make_query(self) -> tuple[str, dict[str, Any]]:
        q = str(
            Object(
                "query",
                {"$owner": "String!", "$page_size": "Int!", "$cursor": "String"},
                Object(
                    "repositoryOwner",
                    {"login": "$owner"},
                    Object(
                        "repositories",
                        {
                            "orderBy": "{field: NAME, direction: ASC}",
                            "first": "$page_size",
                            "after": "$cursor",
                        },
                        Object("nodes", {}, *REPO_FIELDS),
                        Object("pageInfo", {}, "endCursor", "hasNextPage"),
                    ),
                ),
            )
        )
        variables = {
            "owner": self.owner,
            "page_size": PAGE_SIZE,
            "cursor": self.cursor,
        }
        return (q, variables)

    def parse_response(self, data: Any) -> list[Repository]:
        if data["data"]["repositoryOwner"] is None:
            # For some reason, as of 2021-11-01, GitHub handles requests for a
            # nonexistent repositoryOwner by returning `null` instead of
            # responding with a `NOT_FOUND` error like it does for nonexistent
            # users & repositories.
            raise NotFoundError(f"No such repository owner: {self.owner!r}")
        return super().parse_response(data)


@dataclass
class SingleRepoQuery(QueryManager[Repository]):
    owner: str
    name: str

    def make_query(self) -> tuple[str, dict[str, Any]]:
        q = str(
            Object(
                "query",
                {"$owner": "String!", "$name": "String!"},
                Object(
                    "repository", {"owner": "$owner", "name": "$name"}, *REPO_FIELDS
                ),
            )
        )
        return (q, {"owner": self.owner, "name": self.name})

    def parse_response(self, data: Any) -> list[Repository]:
        self.has_next_page = False
        return [Repository.from_node(data["data"]["repository"])]


@dataclass
class ActivityQuery(QueryManager[RepoActivity]):
    repo: Repository
    types: list[ActivityType]
    cursors: CursorDict

    def make_query(self) -> tuple[str, dict[str, Any]]:
        variables: dict[str, Any] = {"repo_id": self.repo.id}
        variable_defs = {"$repo_id": "ID!"}
        connections: list[Object] = []
        for it in self.types:
            if it in self.cursors:
                connections.append(it.event_cls.CONNECTION)
                variable_defs["$page_size"] = "Int!"
                variables["page_size"] = PAGE_SIZE
                variable_defs[f"${it.api_name}_cursor"] = "String"
                variables[f"{it.api_name}_cursor"] = self.cursors[it]
            else:
                connections.append(it.event_cls.LAST_CONNECTION)
        query = Object(
            "query",
            variable_defs,
            Object(
                "node",
                {"id": "$repo_id"},
                Object("... on Repository", {}, *connections),
            ),
        )
        return (str(query), variables)

    def parse_response(self, data: Any) -> list[RepoActivity]:
        events = []
        self.has_next_page = False
        for it in self.types:
            root = data["data"]["node"][it.api_name]
            pageInfo = root.pop("pageInfo")
            new_cursor = pageInfo["endCursor"]
            if it in self.cursors:
                if pageInfo["hasNextPage"]:
                    self.has_next_page = True
                assert len(root) <= 1
                nodes: list[dict] = next(iter(root.values()), [])
                for node in nodes:
                    try:
                        events.append(it.event_cls.from_node(self.repo, node))
                    except BogusEventError as e:
                        log.debug("Discarding bogus event: %s", e)
                if new_cursor is not None:
                    assert isinstance(new_cursor, str)
                    self.cursors[it] = new_cursor
            else:
                self.cursors[it] = new_cursor
        return events



from datetime import datetime, timezone
import pytest
from reponews.types import (
    ActivityType,
    Event,
    NewDiscussionEvent,
    NewForkEvent,
    NewIssueEvent,
    NewPREvent,
    NewReleaseEvent,
    NewStarEvent,
    NewTagEvent,
    RepoActivity,
    RepoRenamedEvent,
    Repository,
    RepoTrackedEvent,
    RepoUntrackedEvent,
    User,
)

REPO = Repository(
    id="id:viewer/repo",
    nameWithOwner="viewer/repo",
    owner=User(
        login="viewer",
        url="https://github.com/viewer",
        name="Vid Ewer",
        isViewer=True,
    ),
    name="repo",
    url="https://github.com/viewer/repo",
    description="My Very Special Repo(tm)",
    descriptionHTML="<div>My Very Special Repo(tm)</div>",
)

USER = User(
    login="luser", url="https://github.com/luser", name="Linux User", isViewer=False
)


def test_all_activity_types() -> None:
    assert list(ActivityType) == [
        ActivityType.ISSUE,
        ActivityType.PR,
        ActivityType.DISCUSSION,
        ActivityType.RELEASE,
        ActivityType.TAG,
        ActivityType.STAR,
        ActivityType.FORK,
    ]


@pytest.mark.parametrize(
    "obj,name,value,api_name,event_cls",
    [
        (ActivityType.ISSUE, "ISSUE", "issue", "issues", NewIssueEvent),
        (ActivityType.PR, "PR", "pr", "pullRequests", NewPREvent),
        (
            ActivityType.DISCUSSION,
            "DISCUSSION",
            "discussion",
            "discussions",
            NewDiscussionEvent,
        ),
        (
            ActivityType.RELEASE,
            "RELEASE",
            "release",
            "releases",
            NewReleaseEvent,
        ),
        (
            ActivityType.TAG,
            "TAG",
            "tag",
            "tags",
            NewTagEvent,
        ),
        (
            ActivityType.STAR,
            "STAR",
            "star",
            "stargazers",
            NewStarEvent,
        ),
        (
            ActivityType.FORK,
            "FORK",
            "fork",
            "forks",
            NewForkEvent,
        ),
    ],
)
def test_activity_type(
    obj: ActivityType, name: str, value: str, api_name: str, event_cls: type
) -> None:
    assert obj.name == name
    assert obj.value == value
    assert obj.api_name == api_name
    assert obj.event_cls is event_cls
    assert obj is ActivityType[name]
    # <https://github.com/python/mypy/issues/10573>
    assert obj is ActivityType(value)  # type: ignore[call-arg]


@pytest.mark.parametrize(
    "ev,s",
    [
        (
            NewIssueEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                number=42,
                title="I found a bug!",
                author=USER,
                url="https://github.com/viewer/repo/issues/42",
            ),
            "[viewer/repo] ISSUE #42: I found a bug! (@luser)\n"
            "<https://github.com/viewer/repo/issues/42>",
        ),
        (
            NewPREvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                number=43,
                title="I fixed the bug",
                author=USER,
                url="https://github.com/viewer/repo/pull/43",
            ),
            "[viewer/repo] PR #43: I fixed the bug (@luser)\n"
            "<https://github.com/viewer/repo/pull/43>",
        ),
        (
            NewDiscussionEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                number=44,
                title="Where do bugs come from?",
                author=USER,
                url="https://github.com/viewer/repo/discussions/44",
            ),
            "[viewer/repo] DISCUSSION #44: Where do bugs come from? (@luser)\n"
            "<https://github.com/viewer/repo/discussions/44>",
        ),
        (
            NewReleaseEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name="v1.0 — Initial release",
                tagName="v1.0.0",
                author=USER,
                description=(
                    "Our first full release!  New in this version:\n"
                    "\n"
                    "- Added a feature\n"
                    "- Fixed a bug"
                ),
                descriptionHTML=(
                    "<p>Our first full release!  New in this version:</p>\n"
                    "<ul>\n"
                    "<li>Added a feature</li>\n"
                    "<li>Fixed a bug</li>\n"
                    "</ul>"
                ),
                isDraft=False,
                isPrerelease=False,
                url="https://github.com/viewer/repo/releases/tag/v1.0.0",
            ),
            "[viewer/repo] RELEASE v1.0.0: v1.0 — Initial release (@luser)\n"
            "<https://github.com/viewer/repo/releases/tag/v1.0.0>\n"
            "> Our first full release!  New in this version:\n"
            "> \n"
            "> - Added a feature\n"
            "> - Fixed a bug",
        ),
        (
            NewReleaseEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name=None,
                tagName="v1.0.0a1",
                author=None,
                description=None,
                descriptionHTML=None,
                isDraft=True,
                isPrerelease=True,
                url="https://github.com/viewer/repo/releases/tag/v1.0.0a1",
            ),
            "[viewer/repo] RELEASE v1.0.0a1 [draft] [prerelease]\n"
            "<https://github.com/viewer/repo/releases/tag/v1.0.0a1>",
        ),
        (
            NewReleaseEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name="Empty description",
                tagName="v1.0.1",
                author=USER,
                description="",
                descriptionHTML="",
                isDraft=False,
                isPrerelease=False,
                url="https://github.com/viewer/repo/releases/tag/v1.0.1",
            ),
            "[viewer/repo] RELEASE v1.0.1: Empty description (@luser)\n"
            "<https://github.com/viewer/repo/releases/tag/v1.0.1>",
        ),
        (
            NewTagEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name="v1.2.3",
                user=USER,
            ),
            "[viewer/repo] TAG v1.2.3 (@luser)\n"
            "<https://github.com/viewer/repo/releases/tag/v1.2.3>",
        ),
        (
            NewTagEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name="v1.2.3",
                user=None,
            ),
            "[viewer/repo] TAG v1.2.3\n"
            "<https://github.com/viewer/repo/releases/tag/v1.2.3>",
        ),
        (
            NewStarEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                user=USER,
            ),
            "★ @luser starred viewer/repo",
        ),
        (
            NewForkEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                fork=Repository(
                    id="id:forker/repo",
                    nameWithOwner="forker/repo",
                    owner=User(
                        login="forker",
                        url="https://github.com/forker",
                        name="Fabian Orker",
                        isViewer=False,
                    ),
                    name="repo",
                    url="https://github.com/forker/repo",
                    description="My Very Special Repo(tm)",
                    descriptionHTML="<div>My Very Special Repo(tm)</div>",
                ),
            ),
            "@forker forked viewer/repo\n<https://github.com/forker/repo>",
        ),
        (
            RepoTrackedEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
            ),
            "Now tracking repository viewer/repo\n"
            "<https://github.com/viewer/repo>\n"
            "> My Very Special Repo(tm)",
        ),
        (
            RepoTrackedEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=Repository(**{**dict(REPO), "description": None}),
            ),
            "Now tracking repository viewer/repo\n<https://github.com/viewer/repo>",
        ),
        (
            RepoUntrackedEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
            ),
            "No longer tracking repository viewer/repo",
        ),
        (
            RepoRenamedEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                old_repo=Repository(
                    id="id:viewer/test",
                    nameWithOwner="viewer/test",
                    owner=User(
                        login="viewer",
                        url="https://github.com/viewer",
                        name="Vid Ewer",
                        isViewer=True,
                    ),
                    name="test",
                    url="https://github.com/viewer/test",
                    description="This is a test.",
                    descriptionHTML="<div>This is a test.</div>",
                ),
            ),
            "Repository renamed: viewer/test → viewer/repo",
        ),
    ],
)
def test_event_render(ev: Event, s: str) -> None:
    assert ev.render() == s


@pytest.mark.parametrize(
    "ev,mine",
    [
        (
            NewIssueEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                number=42,
                title="I found a bug!",
                author=USER,
                url="https://github.com/viewer/repo/issues/42",
            ),
            False,
        ),
        (
            NewPREvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                number=43,
                title="I fixed the bug",
                author=REPO.owner,
                url="https://github.com/viewer/repo/pull/43",
            ),
            True,
        ),
        (
            NewReleaseEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name="v1.0 — Initial release",
                tagName="v1.0.0",
                author=USER,
                description=(
                    "Our first full release!  New in this version:\n"
                    "\n"
                    "- Added a feature\n"
                    "- Fixed a bug"
                ),
                descriptionHTML=(
                    "<p>Our first full release!  New in this version:</p>\n"
                    "<ul>\n"
                    "<li>Added a feature</li>\n"
                    "<li>Fixed a bug</li>\n"
                    "</ul>"
                ),
                isDraft=False,
                isPrerelease=False,
                url="https://github.com/viewer/repo/releases/tag/v1.0.0",
            ),
            False,
        ),
        (
            NewReleaseEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name=None,
                tagName="v1.0.0a1",
                author=None,
                description=None,
                descriptionHTML=None,
                isDraft=True,
                isPrerelease=True,
                url="https://github.com/viewer/repo/releases/tag/v1.0.0a1",
            ),
            False,
        ),
        (
            NewReleaseEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name="Empty description",
                tagName="v1.0.1",
                author=REPO.owner,
                description="",
                descriptionHTML="",
                isDraft=False,
                isPrerelease=False,
                url="https://github.com/viewer/repo/releases/tag/v1.0.1",
            ),
            True,
        ),
        (
            NewTagEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name="v1.2.3",
                user=USER,
            ),
            False,
        ),
        (
            NewTagEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name="v1.2.3",
                user=None,
            ),
            False,
        ),
        (
            NewTagEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                name="v1.2.3",
                user=REPO.owner,
            ),
            True,
        ),
        (
            NewStarEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                user=USER,
            ),
            False,
        ),
        (
            NewStarEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                user=REPO.owner,
            ),
            True,
        ),
        (
            NewForkEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                fork=Repository(
                    id="id:forker/repo",
                    nameWithOwner="forker/repo",
                    owner=User(
                        login="forker",
                        url="https://github.com/forker",
                        name="Fabian Orker",
                        isViewer=False,
                    ),
                    name="repo",
                    url="https://github.com/forker/repo",
                    description="My Very Special Repo(tm)",
                    descriptionHTML="<div>My Very Special Repo(tm)</div>",
                ),
            ),
            False,
        ),
        (
            NewForkEvent(
                timestamp=datetime(2021, 11, 18, 15, 28, 2, tzinfo=timezone.utc),
                repo=REPO,
                fork=Repository(
                    id="id:forker/repo",
                    nameWithOwner="forker/repo",
                    owner=REPO.owner,
                    name="repo",
                    url="https://github.com/forker/repo",
                    description="My Very Special Repo(tm)",
                    descriptionHTML="<div>My Very Special Repo(tm)</div>",
                ),
            ),
            True,
        ),
    ],
)
def test_repo_activity_is_mine(ev: RepoActivity, mine: bool) -> None:
    assert ev.is_mine is mine



from datetime import datetime, timezone
from typing import Optional
import pytest
from reponews.qmanager import (
    PAGE_SIZE,
    ActivityQuery,
    OwnersReposQuery,
    SingleRepoQuery,
    ViewersReposQuery,
)
from reponews.types import (
    ActivityType,
    Affiliation,
    NewDiscussionEvent,
    NewForkEvent,
    NewIssueEvent,
    NewPREvent,
    NewReleaseEvent,
    NewStarEvent,
    NewTagEvent,
    Repository,
    User,
)
from reponews.util import NotFoundError


def test_viewers_repos_query() -> None:
    q = (
        "query(\n"
        "    $page_size: Int!,\n"
        "    $affiliations: [RepositoryAffiliation!],\n"
        "    $cursor: String\n"
        ") {\n"
        "    viewer {\n"
        "        repositories(\n"
        "            ownerAffiliations: $affiliations,\n"
        "            orderBy: {field: NAME, direction: ASC},\n"
        "            first: $page_size,\n"
        "            after: $cursor\n"
        "        ) {\n"
        "            nodes {\n"
        "                id\n"
        "                nameWithOwner\n"
        "                owner {\n"
        "                    login\n"
        "                    url\n"
        "                    ... on User {\n"
        "                        name\n"
        "                        isViewer\n"
        "                    }\n"
        "                    ... on Organization {\n"
        "                        name\n"
        "                    }\n"
        "                }\n"
        "                name\n"
        "                url\n"
        "                description\n"
        "                descriptionHTML\n"
        "            }\n"
        "            pageInfo {\n"
        "                endCursor\n"
        "                hasNextPage\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    manager = ViewersReposQuery(
        affiliations=[Affiliation.OWNER, Affiliation.ORGANIZATION_MEMBER]
    )
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "page_size": PAGE_SIZE,
            "affiliations": ["OWNER", "ORGANIZATION_MEMBER"],
            "cursor": None,
        },
    )
    repos = manager.parse_response(
        {
            "data": {
                "viewer": {
                    "repositories": {
                        "nodes": [
                            {
                                "id": "id:viewer/repo",
                                "nameWithOwner": "viewer/repo",
                                "owner": {
                                    "login": "viewer",
                                    "url": "https://github.com/viewer",
                                    "name": "Vid Ewer",
                                    "isViewer": True,
                                },
                                "name": "repo",
                                "url": "https://github.com/viewer/repo",
                                "description": "My Very Special Repo(tm)",
                                "descriptionHTML": (
                                    "<div>My Very Special Repo(tm)</div>"
                                ),
                            },
                            {
                                "id": "id:viewer/project",
                                "nameWithOwner": "viewer/project",
                                "owner": {
                                    "login": "viewer",
                                    "url": "https://github.com/viewer",
                                    "name": "Vid Ewer",
                                    "isViewer": True,
                                },
                                "name": "project",
                                "url": "https://github.com/viewer/project",
                                "description": "My big ol' project",
                                "descriptionHTML": "<div>My big ol' project</div>",
                            },
                            {
                                "id": "id:org/workspace",
                                "nameWithOwner": "org/workspace",
                                "owner": {
                                    "login": "org",
                                    "url": "https://github.com/org",
                                    "name": "The Org Corp",
                                },
                                "name": "workspace",
                                "url": "https://github.com/org/workspace",
                                "description": "Where we all work",
                                "descriptionHTML": "<div>Where we all work</div>",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:0001",
                            "hasNextPage": True,
                        },
                    }
                }
            }
        }
    )
    assert repos == [
        Repository(
            id="id:viewer/repo",
            nameWithOwner="viewer/repo",
            owner=User(
                login="viewer",
                url="https://github.com/viewer",
                name="Vid Ewer",
                isViewer=True,
            ),
            name="repo",
            url="https://github.com/viewer/repo",
            description="My Very Special Repo(tm)",
            descriptionHTML="<div>My Very Special Repo(tm)</div>",
        ),
        Repository(
            id="id:viewer/project",
            nameWithOwner="viewer/project",
            owner=User(
                login="viewer",
                url="https://github.com/viewer",
                name="Vid Ewer",
                isViewer=True,
            ),
            name="project",
            url="https://github.com/viewer/project",
            description="My big ol' project",
            descriptionHTML="<div>My big ol' project</div>",
        ),
        Repository(
            id="id:org/workspace",
            nameWithOwner="org/workspace",
            owner=User(
                login="org",
                url="https://github.com/org",
                name="The Org Corp",
                isViewer=False,
            ),
            name="workspace",
            url="https://github.com/org/workspace",
            description="Where we all work",
            descriptionHTML="<div>Where we all work</div>",
        ),
    ]
    assert manager.cursor == "cursor:0001"
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "page_size": PAGE_SIZE,
            "affiliations": ["OWNER", "ORGANIZATION_MEMBER"],
            "cursor": "cursor:0001",
        },
    )
    repos = manager.parse_response(
        {
            "data": {
                "viewer": {
                    "repositories": {
                        "nodes": [
                            {
                                "id": "id:viewer/test",
                                "nameWithOwner": "viewer/test",
                                "owner": {
                                    "login": "viewer",
                                    "url": "https://github.com/viewer",
                                    "name": "Vid Ewer",
                                    "isViewer": True,
                                },
                                "name": "test",
                                "url": "https://github.com/viewer/test",
                                "description": None,
                                "descriptionHTML": "<div></div>",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:0002",
                            "hasNextPage": False,
                        },
                    }
                }
            }
        }
    )
    assert repos == [
        Repository(
            id="id:viewer/test",
            nameWithOwner="viewer/test",
            owner=User(
                login="viewer",
                url="https://github.com/viewer",
                name="Vid Ewer",
                isViewer=True,
            ),
            name="test",
            url="https://github.com/viewer/test",
            description=None,
            descriptionHTML="<div></div>",
        )
    ]
    assert manager.cursor == "cursor:0002"
    assert not manager.has_next_page


def test_owners_repos_query() -> None:
    q = (
        "query(\n"
        "    $owner: String!,\n"
        "    $page_size: Int!,\n"
        "    $cursor: String\n"
        ") {\n"
        "    repositoryOwner(login: $owner) {\n"
        "        repositories(\n"
        "            orderBy: {field: NAME, direction: ASC},\n"
        "            first: $page_size,\n"
        "            after: $cursor\n"
        "        ) {\n"
        "            nodes {\n"
        "                id\n"
        "                nameWithOwner\n"
        "                owner {\n"
        "                    login\n"
        "                    url\n"
        "                    ... on User {\n"
        "                        name\n"
        "                        isViewer\n"
        "                    }\n"
        "                    ... on Organization {\n"
        "                        name\n"
        "                    }\n"
        "                }\n"
        "                name\n"
        "                url\n"
        "                description\n"
        "                descriptionHTML\n"
        "            }\n"
        "            pageInfo {\n"
        "                endCursor\n"
        "                hasNextPage\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    manager = OwnersReposQuery(owner="repo.owner")
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {"owner": "repo.owner", "page_size": PAGE_SIZE, "cursor": None},
    )
    repos = manager.parse_response(
        {
            "data": {
                "repositoryOwner": {
                    "repositories": {
                        "nodes": [
                            {
                                "id": "id:repo.owner/repo",
                                "nameWithOwner": "repo.owner/repo",
                                "owner": {
                                    "login": "repo.owner",
                                    "url": "https://github.com/repo.owner",
                                    "name": "Repository Owner",
                                    "isViewer": False,
                                },
                                "name": "repo",
                                "url": "https://github.com/repo.owner/repo",
                                "description": "My Very Special Repo(tm)",
                                "descriptionHTML": (
                                    "<div>My Very Special Repo(tm)</div>"
                                ),
                            },
                            {
                                "id": "id:repo.owner/project",
                                "nameWithOwner": "repo.owner/project",
                                "owner": {
                                    "login": "repo.owner",
                                    "url": "https://github.com/repo.owner",
                                    "name": "Repository Owner",
                                    "isViewer": False,
                                },
                                "name": "project",
                                "url": "https://github.com/repo.owner/project",
                                "description": "My big ol' project",
                                "descriptionHTML": "<div>My big ol' project</div>",
                            },
                            {
                                "id": "id:org/workspace",
                                "nameWithOwner": "org/workspace",
                                "owner": {
                                    "login": "org",
                                    "url": "https://github.com/org",
                                    "name": "The Org Corp",
                                },
                                "name": "workspace",
                                "url": "https://github.com/org/workspace",
                                "description": "Where we all work",
                                "descriptionHTML": "<div>Where we all work</div>",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:0001",
                            "hasNextPage": True,
                        },
                    }
                }
            }
        }
    )
    assert repos == [
        Repository(
            id="id:repo.owner/repo",
            nameWithOwner="repo.owner/repo",
            owner=User(
                login="repo.owner",
                url="https://github.com/repo.owner",
                name="Repository Owner",
                isViewer=False,
            ),
            name="repo",
            url="https://github.com/repo.owner/repo",
            description="My Very Special Repo(tm)",
            descriptionHTML="<div>My Very Special Repo(tm)</div>",
        ),
        Repository(
            id="id:repo.owner/project",
            nameWithOwner="repo.owner/project",
            owner=User(
                login="repo.owner",
                url="https://github.com/repo.owner",
                name="Repository Owner",
                isViewer=False,
            ),
            name="project",
            url="https://github.com/repo.owner/project",
            description="My big ol' project",
            descriptionHTML="<div>My big ol' project</div>",
        ),
        Repository(
            id="id:org/workspace",
            nameWithOwner="org/workspace",
            owner=User(
                login="org",
                url="https://github.com/org",
                name="The Org Corp",
                isViewer=False,
            ),
            name="workspace",
            url="https://github.com/org/workspace",
            description="Where we all work",
            descriptionHTML="<div>Where we all work</div>",
        ),
    ]
    assert manager.cursor == "cursor:0001"
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {"owner": "repo.owner", "page_size": PAGE_SIZE, "cursor": "cursor:0001"},
    )
    repos = manager.parse_response(
        {
            "data": {
                "repositoryOwner": {
                    "repositories": {
                        "nodes": [
                            {
                                "id": "id:repo.owner/test",
                                "nameWithOwner": "repo.owner/test",
                                "owner": {
                                    "login": "repo.owner",
                                    "url": "https://github.com/repo.owner",
                                    "name": "Repository Owner",
                                    "isViewer": False,
                                },
                                "name": "test",
                                "url": "https://github.com/repo.owner/test",
                                "description": None,
                                "descriptionHTML": "<div></div>",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:0002",
                            "hasNextPage": False,
                        },
                    }
                }
            }
        }
    )
    assert repos == [
        Repository(
            id="id:repo.owner/test",
            nameWithOwner="repo.owner/test",
            owner=User(
                login="repo.owner",
                url="https://github.com/repo.owner",
                name="Repository Owner",
                isViewer=False,
            ),
            name="test",
            url="https://github.com/repo.owner/test",
            description=None,
            descriptionHTML="<div></div>",
        )
    ]
    assert manager.cursor == "cursor:0002"
    assert not manager.has_next_page


def test_owners_repos_query_not_found() -> None:
    manager = OwnersReposQuery(owner="DNE")
    with pytest.raises(NotFoundError):
        manager.parse_response({"data": {"repositoryOwner": None}})


def test_owners_repos_query_no_repos() -> None:
    manager = OwnersReposQuery(owner="newbie")
    assert (
        manager.parse_response(
            {
                "data": {
                    "repositoryOwner": {
                        "repositories": {
                            "nodes": [],
                            "pageInfo": {
                                "endCursor": None,
                                "hasNextPage": False,
                            },
                        }
                    }
                }
            }
        )
        == []
    )
    assert manager.cursor is None
    assert not manager.has_next_page


def test_single_repo_query() -> None:
    q = (
        "query($owner: String!, $name: String!) {\n"
        "    repository(owner: $owner, name: $name) {\n"
        "        id\n"
        "        nameWithOwner\n"
        "        owner {\n"
        "            login\n"
        "            url\n"
        "            ... on User {\n"
        "                name\n"
        "                isViewer\n"
        "            }\n"
        "            ... on Organization {\n"
        "                name\n"
        "            }\n"
        "        }\n"
        "        name\n"
        "        url\n"
        "        description\n"
        "        descriptionHTML\n"
        "    }\n"
        "}\n"
    )
    manager = SingleRepoQuery(owner="repo.owner", name="repo")
    assert manager.has_next_page
    assert manager.make_query() == (q, {"owner": "repo.owner", "name": "repo"})
    repos = manager.parse_response(
        {
            "data": {
                "repository": {
                    "id": "id:repo.owner/repo",
                    "nameWithOwner": "repo.owner/repo",
                    "owner": {
                        "login": "repo.owner",
                        "url": "https://github.com/repo.owner",
                        "name": "Repository Owner",
                        "isViewer": False,
                    },
                    "name": "repo",
                    "url": "https://github.com/repo.owner/repo",
                    "description": "My Very Special Repo(tm)",
                    "descriptionHTML": "<div>My Very Special Repo(tm)</div>",
                },
            }
        }
    )
    assert repos == [
        Repository(
            id="id:repo.owner/repo",
            nameWithOwner="repo.owner/repo",
            owner=User(
                login="repo.owner",
                url="https://github.com/repo.owner",
                name="Repository Owner",
                isViewer=False,
            ),
            name="repo",
            url="https://github.com/repo.owner/repo",
            description="My Very Special Repo(tm)",
            descriptionHTML="<div>My Very Special Repo(tm)</div>",
        ),
    ]
    assert not manager.has_next_page


def test_activity_query() -> None:
    q = (
        "query(\n"
        "    $repo_id: ID!,\n"
        "    $page_size: Int!,\n"
        "    $pullRequests_cursor: String\n"
        ") {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            pullRequests(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $pullRequests_cursor\n"
        "            ) {\n"
        "                nodes {\n"
        "                    author {\n"
        "                        login\n"
        "                        url\n"
        "                        ... on User {\n"
        "                            name\n"
        "                            isViewer\n"
        "                        }\n"
        "                    }\n"
        "                    createdAt\n"
        "                    number\n"
        "                    title\n"
        "                    url\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(
        repo=repo, types=[ActivityType.PR], cursors={ActivityType.PR: "cursor:0001"}
    )
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "pullRequests_cursor": "cursor:0001",
        },
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "pullRequests": {
                        "nodes": [
                            {
                                "author": {
                                    "login": "a.contributor",
                                    "url": "https://github.com/a.contributor",
                                    "name": "A. Contributor",
                                    "isViewer": False,
                                },
                                "createdAt": "2021-07-04T12:15:07Z",
                                "number": 1,
                                "title": "Add a feature",
                                "url": "https://github.com/viewer/repo/pull/1",
                            },
                            {
                                "author": {
                                    "login": "prbot",
                                    "url": "https://github.com/apps/prbot",
                                },
                                "createdAt": "2021-07-04T12:34:56Z",
                                "number": 2,
                                "title": "Automated pull request",
                                "url": "https://github.com/viewer/repo/pull/2",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:0002",
                            "hasNextPage": True,
                        },
                    }
                }
            }
        }
    )
    assert events == [
        NewPREvent(
            timestamp=datetime(2021, 7, 4, 12, 15, 7, tzinfo=timezone.utc),
            repo=repo,
            number=1,
            title="Add a feature",
            url="https://github.com/viewer/repo/pull/1",
            author=User(
                login="a.contributor",
                url="https://github.com/a.contributor",
                name="A. Contributor",
                isViewer=False,
            ),
        ),
        NewPREvent(
            timestamp=datetime(2021, 7, 4, 12, 34, 56, tzinfo=timezone.utc),
            repo=repo,
            number=2,
            title="Automated pull request",
            url="https://github.com/viewer/repo/pull/2",
            author=User(
                login="prbot",
                url="https://github.com/apps/prbot",
                name=None,
                isViewer=False,
            ),
        ),
    ]
    assert manager.cursors == {ActivityType.PR: "cursor:0002"}
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "pullRequests_cursor": "cursor:0002",
        },
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "pullRequests": {
                        "nodes": [
                            {
                                "author": {
                                    "login": "new-user",
                                    "url": "https://github.com/new-user",
                                    "name": None,
                                    "isViewer": False,
                                },
                                "createdAt": "2021-07-05T01:02:03Z",
                                "number": 4,
                                "title": "What am I doing?",
                                "url": "https://github.com/viewer/repo/pull/4",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:0003",
                            "hasNextPage": False,
                        },
                    }
                }
            }
        }
    )
    assert events == [
        NewPREvent(
            timestamp=datetime(2021, 7, 5, 1, 2, 3, tzinfo=timezone.utc),
            repo=repo,
            number=4,
            title="What am I doing?",
            url="https://github.com/viewer/repo/pull/4",
            author=User(
                login="new-user",
                url="https://github.com/new-user",
                name=None,
                isViewer=False,
            ),
        ),
    ]
    assert manager.cursors == {ActivityType.PR: "cursor:0003"}
    assert not manager.has_next_page


def test_activity_query_no_events() -> None:
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(
        repo=repo, types=[ActivityType.PR], cursors={ActivityType.PR: "cursor:0001"}
    )
    assert (
        manager.parse_response(
            {
                "data": {
                    "node": {
                        "pullRequests": {
                            "nodes": [],
                            "pageInfo": {
                                "endCursor": None,
                                "hasNextPage": False,
                            },
                        }
                    }
                }
            }
        )
        == []
    )
    assert manager.cursors == {ActivityType.PR: "cursor:0001"}
    assert not manager.has_next_page


def test_activity_query_null_cursor_no_events() -> None:
    q = (
        "query(\n"
        "    $repo_id: ID!,\n"
        "    $page_size: Int!,\n"
        "    $discussions_cursor: String\n"
        ") {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            discussions(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $discussions_cursor\n"
        "            ) {\n"
        "                nodes {\n"
        "                    author {\n"
        "                        login\n"
        "                        url\n"
        "                        ... on User {\n"
        "                            name\n"
        "                            isViewer\n"
        "                        }\n"
        "                    }\n"
        "                    createdAt\n"
        "                    number\n"
        "                    title\n"
        "                    url\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(
        repo=repo,
        types=[ActivityType.DISCUSSION],
        cursors={ActivityType.DISCUSSION: None},
    )
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "discussions_cursor": None,
        },
    )
    assert (
        manager.parse_response(
            {
                "data": {
                    "node": {
                        "discussions": {
                            "nodes": [],
                            "pageInfo": {
                                "endCursor": None,
                                "hasNextPage": False,
                            },
                        }
                    }
                }
            }
        )
        == []
    )
    assert manager.cursors == {ActivityType.DISCUSSION: None}
    assert not manager.has_next_page


def test_activity_query_null_cursor_some_events() -> None:
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(
        repo=repo,
        types=[ActivityType.DISCUSSION],
        cursors={ActivityType.DISCUSSION: None},
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "discussions": {
                        "nodes": [
                            {
                                "author": {
                                    "login": "need-to-know",
                                    "url": "https://github.com/need-to-know",
                                    "name": "Please tell me",
                                    "isViewer": False,
                                },
                                "createdAt": "2021-11-04T19:46:29Z",
                                "number": 42,
                                "title": "How is everybody doing?",
                                "url": "https://github.com/viewer/repo/discussions/42",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:d001",
                            "hasNextPage": False,
                        },
                    }
                }
            }
        }
    )
    assert events == [
        NewDiscussionEvent(
            timestamp=datetime(2021, 11, 4, 19, 46, 29, tzinfo=timezone.utc),
            repo=repo,
            number=42,
            title="How is everybody doing?",
            url="https://github.com/viewer/repo/discussions/42",
            author=User(
                login="need-to-know",
                url="https://github.com/need-to-know",
                name="Please tell me",
                isViewer=False,
            ),
        ),
    ]
    assert manager.cursors == {ActivityType.DISCUSSION: "cursor:d001"}
    assert not manager.has_next_page


@pytest.mark.parametrize("new_cursor", [None, "cursor:0001"])
def test_activity_query_no_cursor(new_cursor: Optional[str]) -> None:
    q = (
        "query($repo_id: ID!) {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            discussions(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                last: 1\n"
        "            ) {\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(repo=repo, types=[ActivityType.DISCUSSION], cursors={})
    assert manager.has_next_page
    assert manager.make_query() == (q, {"repo_id": "id:viewer/repo"})
    assert (
        manager.parse_response(
            {
                "data": {
                    "node": {
                        "discussions": {
                            "pageInfo": {"endCursor": new_cursor},
                        }
                    }
                }
            }
        )
        == []
    )
    assert manager.cursors == {ActivityType.DISCUSSION: new_cursor}
    assert not manager.has_next_page


def test_activity_query_multiple_types() -> None:
    q = (
        "query(\n"
        "    $repo_id: ID!,\n"
        "    $page_size: Int!,\n"
        "    $issues_cursor: String,\n"
        "    $discussions_cursor: String\n"
        ") {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            issues(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $issues_cursor\n"
        "            ) {\n"
        "                nodes {\n"
        "                    author {\n"
        "                        login\n"
        "                        url\n"
        "                        ... on User {\n"
        "                            name\n"
        "                            isViewer\n"
        "                        }\n"
        "                    }\n"
        "                    createdAt\n"
        "                    number\n"
        "                    title\n"
        "                    url\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "            discussions(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $discussions_cursor\n"
        "            ) {\n"
        "                nodes {\n"
        "                    author {\n"
        "                        login\n"
        "                        url\n"
        "                        ... on User {\n"
        "                            name\n"
        "                            isViewer\n"
        "                        }\n"
        "                    }\n"
        "                    createdAt\n"
        "                    number\n"
        "                    title\n"
        "                    url\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(
        repo=repo,
        types=[ActivityType.ISSUE, ActivityType.DISCUSSION],
        cursors={
            ActivityType.ISSUE: "cursor:i001",
            ActivityType.DISCUSSION: "cursor:d001",
        },
    )
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "issues_cursor": "cursor:i001",
            "discussions_cursor": "cursor:d001",
        },
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "issues": {
                        "nodes": [
                            {
                                "author": {
                                    "login": "bfg",
                                    "url": "https://github.com/bfg",
                                    "name": "Bug Finder General",
                                    "isViewer": False,
                                },
                                "createdAt": "2021-07-04T12:15:07Z",
                                "number": 1,
                                "title": "Found a bug",
                                "url": "https://github.com/viewer/repo/issues/1",
                            },
                            {
                                "author": {
                                    "login": "needstuff",
                                    "url": "https://github.com/needstuff",
                                    "name": None,
                                    "isViewer": False,
                                },
                                "createdAt": "2021-07-04T12:34:56Z",
                                "number": 2,
                                "title": "I would like to request a feature",
                                "url": "https://github.com/viewer/repo/issues/2",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:i002",
                            "hasNextPage": True,
                        },
                    },
                    "discussions": {
                        "nodes": [
                            {
                                "author": {
                                    "login": "repo-therapist",
                                    "url": "https://github.com/repo-therapist",
                                    "name": "Repository Therapy",
                                    "isViewer": False,
                                },
                                "createdAt": "2021-07-05T01:02:03Z",
                                "number": 3,
                                "title": "Let's talk about your code.",
                                "url": "https://github.com/viewer/repo/discussions/3",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:d002",
                            "hasNextPage": False,
                        },
                    },
                }
            }
        }
    )
    assert events == [
        NewIssueEvent(
            timestamp=datetime(2021, 7, 4, 12, 15, 7, tzinfo=timezone.utc),
            repo=repo,
            number=1,
            title="Found a bug",
            url="https://github.com/viewer/repo/issues/1",
            author=User(
                login="bfg",
                url="https://github.com/bfg",
                name="Bug Finder General",
                isViewer=False,
            ),
        ),
        NewIssueEvent(
            timestamp=datetime(2021, 7, 4, 12, 34, 56, tzinfo=timezone.utc),
            repo=repo,
            number=2,
            title="I would like to request a feature",
            url="https://github.com/viewer/repo/issues/2",
            author=User(
                login="needstuff",
                url="https://github.com/needstuff",
                name=None,
                isViewer=False,
            ),
        ),
        NewDiscussionEvent(
            timestamp=datetime(2021, 7, 5, 1, 2, 3, tzinfo=timezone.utc),
            repo=repo,
            number=3,
            title="Let's talk about your code.",
            url="https://github.com/viewer/repo/discussions/3",
            author=User(
                login="repo-therapist",
                url="https://github.com/repo-therapist",
                name="Repository Therapy",
                isViewer=False,
            ),
        ),
    ]
    assert manager.cursors == {
        ActivityType.ISSUE: "cursor:i002",
        ActivityType.DISCUSSION: "cursor:d002",
    }
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "issues_cursor": "cursor:i002",
            "discussions_cursor": "cursor:d002",
        },
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "issues": {
                        "nodes": [
                            {
                                "author": {
                                    "login": "new-user",
                                    "url": "https://github.com/new-user",
                                    "name": None,
                                    "isViewer": False,
                                },
                                "createdAt": "2021-07-06T01:02:03Z",
                                "number": 4,
                                "title": "What am I doing?",
                                "url": "https://github.com/viewer/repo/issues/4",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:i003",
                            "hasNextPage": False,
                        },
                    },
                    "discussions": {
                        "nodes": [],
                        "pageInfo": {
                            "endCursor": None,
                            "hasNextPage": False,
                        },
                    },
                }
            }
        }
    )
    assert events == [
        NewIssueEvent(
            timestamp=datetime(2021, 7, 6, 1, 2, 3, tzinfo=timezone.utc),
            repo=repo,
            number=4,
            title="What am I doing?",
            url="https://github.com/viewer/repo/issues/4",
            author=User(
                login="new-user",
                url="https://github.com/new-user",
                name=None,
                isViewer=False,
            ),
        ),
    ]
    assert manager.cursors == {
        ActivityType.ISSUE: "cursor:i003",
        ActivityType.DISCUSSION: "cursor:d002",
    }
    assert not manager.has_next_page


@pytest.mark.parametrize("new_cursor", [None, "cursor:d001"])
def test_activity_query_multiple_types_one_unset_cursor(
    new_cursor: Optional[str],
) -> None:
    q1 = (
        "query(\n"
        "    $repo_id: ID!,\n"
        "    $page_size: Int!,\n"
        "    $issues_cursor: String\n"
        ") {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            issues(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $issues_cursor\n"
        "            ) {\n"
        "                nodes {\n"
        "                    author {\n"
        "                        login\n"
        "                        url\n"
        "                        ... on User {\n"
        "                            name\n"
        "                            isViewer\n"
        "                        }\n"
        "                    }\n"
        "                    createdAt\n"
        "                    number\n"
        "                    title\n"
        "                    url\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "            discussions(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                last: 1\n"
        "            ) {\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )

    q2 = (
        "query(\n"
        "    $repo_id: ID!,\n"
        "    $page_size: Int!,\n"
        "    $issues_cursor: String,\n"
        "    $discussions_cursor: String\n"
        ") {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            issues(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $issues_cursor\n"
        "            ) {\n"
        "                nodes {\n"
        "                    author {\n"
        "                        login\n"
        "                        url\n"
        "                        ... on User {\n"
        "                            name\n"
        "                            isViewer\n"
        "                        }\n"
        "                    }\n"
        "                    createdAt\n"
        "                    number\n"
        "                    title\n"
        "                    url\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "            discussions(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $discussions_cursor\n"
        "            ) {\n"
        "                nodes {\n"
        "                    author {\n"
        "                        login\n"
        "                        url\n"
        "                        ... on User {\n"
        "                            name\n"
        "                            isViewer\n"
        "                        }\n"
        "                    }\n"
        "                    createdAt\n"
        "                    number\n"
        "                    title\n"
        "                    url\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )

    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(
        repo=repo,
        types=[ActivityType.ISSUE, ActivityType.DISCUSSION],
        cursors={ActivityType.ISSUE: "cursor:i001"},
    )
    assert manager.has_next_page
    assert manager.make_query() == (
        q1,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "issues_cursor": "cursor:i001",
        },
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "issues": {
                        "nodes": [
                            {
                                "author": {
                                    "login": "bfg",
                                    "url": "https://github.com/bfg",
                                    "name": "Bug Finder General",
                                    "isViewer": False,
                                },
                                "createdAt": "2021-07-04T12:15:07Z",
                                "number": 1,
                                "title": "Found a bug",
                                "url": "https://github.com/viewer/repo/issues/1",
                            },
                            {
                                "author": {
                                    "login": "needstuff",
                                    "url": "https://github.com/needstuff",
                                    "name": None,
                                    "isViewer": False,
                                },
                                "createdAt": "2021-07-04T12:34:56Z",
                                "number": 2,
                                "title": "I would like to request a feature",
                                "url": "https://github.com/viewer/repo/issues/2",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:i002",
                            "hasNextPage": True,
                        },
                    },
                    "discussions": {
                        "pageInfo": {"endCursor": new_cursor},
                    },
                }
            }
        }
    )
    assert events == [
        NewIssueEvent(
            timestamp=datetime(2021, 7, 4, 12, 15, 7, tzinfo=timezone.utc),
            repo=repo,
            number=1,
            title="Found a bug",
            url="https://github.com/viewer/repo/issues/1",
            author=User(
                login="bfg",
                url="https://github.com/bfg",
                name="Bug Finder General",
                isViewer=False,
            ),
        ),
        NewIssueEvent(
            timestamp=datetime(2021, 7, 4, 12, 34, 56, tzinfo=timezone.utc),
            repo=repo,
            number=2,
            title="I would like to request a feature",
            url="https://github.com/viewer/repo/issues/2",
            author=User(
                login="needstuff",
                url="https://github.com/needstuff",
                name=None,
                isViewer=False,
            ),
        ),
    ]
    assert manager.cursors == {
        ActivityType.ISSUE: "cursor:i002",
        ActivityType.DISCUSSION: new_cursor,
    }
    assert manager.has_next_page
    assert manager.make_query() == (
        q2,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "issues_cursor": "cursor:i002",
            "discussions_cursor": new_cursor,
        },
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "issues": {
                        "nodes": [
                            {
                                "author": {
                                    "login": "new-user",
                                    "url": "https://github.com/new-user",
                                    "name": None,
                                    "isViewer": False,
                                },
                                "createdAt": "2021-07-06T01:02:03Z",
                                "number": 4,
                                "title": "What am I doing?",
                                "url": "https://github.com/viewer/repo/issues/4",
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:i003",
                            "hasNextPage": False,
                        },
                    },
                    "discussions": {
                        "nodes": [],
                        "pageInfo": {
                            "endCursor": None,
                            "hasNextPage": False,
                        },
                    },
                }
            }
        }
    )
    assert events == [
        NewIssueEvent(
            timestamp=datetime(2021, 7, 6, 1, 2, 3, tzinfo=timezone.utc),
            repo=repo,
            number=4,
            title="What am I doing?",
            url="https://github.com/viewer/repo/issues/4",
            author=User(
                login="new-user",
                url="https://github.com/new-user",
                name=None,
                isViewer=False,
            ),
        ),
    ]
    assert manager.cursors == {
        ActivityType.ISSUE: "cursor:i003",
        ActivityType.DISCUSSION: new_cursor,
    }
    assert not manager.has_next_page


def test_activity_query_releases() -> None:
    q = (
        "query(\n"
        "    $repo_id: ID!,\n"
        "    $page_size: Int!,\n"
        "    $releases_cursor: String\n"
        ") {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            releases(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $releases_cursor\n"
        "            ) {\n"
        "                nodes {\n"
        "                    name\n"
        "                    tagName\n"
        "                    author {\n"
        "                        login\n"
        "                        url\n"
        "                        name\n"
        "                        isViewer\n"
        "                    }\n"
        "                    createdAt\n"
        "                    description\n"
        "                    descriptionHTML\n"
        "                    isPrerelease\n"
        "                    isDraft\n"
        "                    url\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(
        repo=repo,
        types=[ActivityType.RELEASE],
        cursors={ActivityType.RELEASE: "cursor:0001"},
    )
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "releases_cursor": "cursor:0001",
        },
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "releases": {
                        "nodes": [
                            {
                                "name": "v0.1.0 — Initial release",
                                "tagName": "v0.1.0",
                                "author": {
                                    "login": "viewer",
                                    "url": "https://github.com/viewer",
                                    "name": "Vid Ewer",
                                    "isViewer": True,
                                },
                                "createdAt": "2021-11-17T15:15:57Z",
                                "description": None,
                                "descriptionHTML": None,
                                "isPrerelease": False,
                                "isDraft": False,
                                "url": (
                                    "https://github.com/viewer/repo/releases/tag/v0.1.0"
                                ),
                            },
                            {
                                "name": "v0.1.0 — Bug fix",
                                "tagName": "v0.1.1",
                                "author": None,
                                "createdAt": "2021-11-17T16:12:34Z",
                                "description": (
                                    "* Fixed a bug in the thingy.\r\n"
                                    "* Hopefully didn't introduce any new bugs"
                                ),
                                "descriptionHTML": (
                                    "<ul>\n"
                                    "<li>Fixed a bug in the thingy.</li>\n"
                                    "<li>Hopefully didn't introduce any new bugs</li>\n"
                                    "</ul>"
                                ),
                                "isPrerelease": False,
                                "isDraft": False,
                                "url": (
                                    "https://github.com/viewer/repo/releases/tag/v0.1.1"
                                ),
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:0002",
                            "hasNextPage": False,
                        },
                    },
                }
            }
        }
    )
    assert events == [
        NewReleaseEvent(
            timestamp=datetime(2021, 11, 17, 15, 15, 57, tzinfo=timezone.utc),
            repo=repo,
            name="v0.1.0 — Initial release",
            tagName="v0.1.0",
            author=User(
                login="viewer",
                url="https://github.com/viewer",
                name="Vid Ewer",
                isViewer=True,
            ),
            description=None,
            descriptionHTML=None,
            isPrerelease=False,
            isDraft=False,
            url="https://github.com/viewer/repo/releases/tag/v0.1.0",
        ),
        NewReleaseEvent(
            timestamp=datetime(2021, 11, 17, 16, 12, 34, tzinfo=timezone.utc),
            repo=repo,
            name="v0.1.0 — Bug fix",
            tagName="v0.1.1",
            author=None,
            description=(
                "* Fixed a bug in the thingy.\n"
                "* Hopefully didn't introduce any new bugs"
            ),
            descriptionHTML=(
                "<ul>\n"
                "<li>Fixed a bug in the thingy.</li>\n"
                "<li>Hopefully didn't introduce any new bugs</li>\n"
                "</ul>"
            ),
            isPrerelease=False,
            isDraft=False,
            url="https://github.com/viewer/repo/releases/tag/v0.1.1",
        ),
    ]
    assert manager.cursors == {ActivityType.RELEASE: "cursor:0002"}
    assert not manager.has_next_page


def test_activity_query_releases_no_cursor() -> None:
    q = (
        "query($repo_id: ID!) {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            releases(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                last: 1\n"
        "            ) {\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(repo=repo, types=[ActivityType.RELEASE], cursors={})
    assert manager.has_next_page
    assert manager.make_query() == (q, {"repo_id": "id:viewer/repo"})


def test_activity_query_tags() -> None:
    q = (
        "query(\n"
        "    $repo_id: ID!,\n"
        "    $page_size: Int!,\n"
        "    $tags_cursor: String\n"
        ") {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            tags: refs(\n"
        '                refPrefix: "refs/tags/",\n'
        "                orderBy: {field: TAG_COMMIT_DATE, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $tags_cursor\n"
        "            ) {\n"
        "                nodes {\n"
        "                    name\n"
        "                    target {\n"
        "                        __typename\n"
        "                        ... on Tag {\n"
        "                            tagger {\n"
        "                                date\n"
        "                                user {\n"
        "                                    login\n"
        "                                    url\n"
        "                                    name\n"
        "                                    isViewer\n"
        "                                }\n"
        "                            }\n"
        "                        }\n"
        "                        ... on Commit {\n"
        "                            committedDate\n"
        "                            author {\n"
        "                                user {\n"
        "                                    login\n"
        "                                    url\n"
        "                                    name\n"
        "                                    isViewer\n"
        "                                }\n"
        "                            }\n"
        "                        }\n"
        "                    }\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(
        repo=repo,
        types=[ActivityType.TAG],
        cursors={ActivityType.TAG: "cursor:0001"},
    )
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "tags_cursor": "cursor:0001",
        },
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "tags": {
                        "nodes": [
                            {
                                "name": "v0.0.0",
                                "target": {
                                    "__typename": "Commit",
                                    "committedDate": "2021-11-17T15:56:48Z",
                                    "author": {
                                        "user": {
                                            "login": "viewer",
                                            "url": "https://github.com/viewer",
                                            "name": "Vid Ewer",
                                            "isViewer": True,
                                        },
                                    },
                                },
                            },
                            {
                                "name": "v0.0.0a0",
                                "target": {
                                    "__typename": "Commit",
                                    "committedDate": "2021-11-17T15:56:50Z",
                                    "author": None,
                                },
                            },
                            {
                                "name": "v0.1.0",
                                "target": {
                                    "__typename": "Tag",
                                    "tagger": {
                                        "date": "2021-11-17T16:04:31Z",
                                        "user": {
                                            "login": "viewer",
                                            "url": "https://github.com/viewer",
                                            "name": "Vid Ewer",
                                            "isViewer": True,
                                        },
                                    },
                                },
                            },
                            {
                                "name": "v0.1.0-tree",
                                "target": {"__typename": "Tree"},
                            },
                            {
                                "name": "v0.0.1a",
                                "target": {
                                    "__typename": "Commit",
                                    "committedDate": None,
                                    "author": None,
                                },
                            },
                            {
                                "name": "v0.1.1",
                                "target": {"__typename": "Tag", "tagger": None},
                            },
                            {
                                "name": "v0.1.2",
                                "target": {
                                    "__typename": "Tag",
                                    "tagger": {
                                        "date": "2021-11-18T00:00:00Z",
                                        "user": None,
                                    },
                                },
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:0002",
                            "hasNextPage": False,
                        },
                    },
                }
            }
        }
    )
    assert events == [
        NewTagEvent(
            timestamp=datetime(2021, 11, 17, 15, 56, 48, tzinfo=timezone.utc),
            repo=repo,
            name="v0.0.0",
            user=User(
                login="viewer",
                url="https://github.com/viewer",
                name="Vid Ewer",
                isViewer=True,
            ),
        ),
        NewTagEvent(
            timestamp=datetime(2021, 11, 17, 15, 56, 50, tzinfo=timezone.utc),
            repo=repo,
            name="v0.0.0a0",
            user=None,
        ),
        NewTagEvent(
            timestamp=datetime(2021, 11, 17, 16, 4, 31, tzinfo=timezone.utc),
            repo=repo,
            name="v0.1.0",
            user=User(
                login="viewer",
                url="https://github.com/viewer",
                name="Vid Ewer",
                isViewer=True,
            ),
        ),
        NewTagEvent(
            timestamp=datetime(2021, 11, 18, 0, 0, 0, tzinfo=timezone.utc),
            repo=repo,
            name="v0.1.2",
            user=None,
        ),
    ]
    assert manager.cursors == {ActivityType.TAG: "cursor:0002"}
    assert not manager.has_next_page


def test_activity_query_tags_no_cursor() -> None:
    q = (
        "query($repo_id: ID!) {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            tags: refs(\n"
        '                refPrefix: "refs/tags/",\n'
        "                orderBy: {field: TAG_COMMIT_DATE, direction: ASC},\n"
        "                last: 1\n"
        "            ) {\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(repo=repo, types=[ActivityType.TAG], cursors={})
    assert manager.has_next_page
    assert manager.make_query() == (q, {"repo_id": "id:viewer/repo"})


def test_activity_query_stargazers() -> None:
    q = (
        "query(\n"
        "    $repo_id: ID!,\n"
        "    $page_size: Int!,\n"
        "    $stargazers_cursor: String\n"
        ") {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            stargazers(\n"
        "                orderBy: {field: STARRED_AT, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $stargazers_cursor\n"
        "            ) {\n"
        "                edges {\n"
        "                    starredAt\n"
        "                    user: node {\n"
        "                        login\n"
        "                        url\n"
        "                        name\n"
        "                        isViewer\n"
        "                    }\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(
        repo=repo,
        types=[ActivityType.STAR],
        cursors={ActivityType.STAR: "cursor:0001"},
    )
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "stargazers_cursor": "cursor:0001",
        },
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "stargazers": {
                        "edges": [
                            {
                                "starredAt": "2021-11-17T15:26:39Z",
                                "user": {
                                    "login": "repo.fan",
                                    "url": "https://github.com/repo.fan",
                                    "name": "Fan of Repositories",
                                    "isViewer": False,
                                },
                            },
                            {
                                "starredAt": "2021-11-17T23:45:56Z",
                                "user": {
                                    "login": "starry.user",
                                    "url": "https://github.com/starry.user",
                                    "name": None,
                                    "isViewer": False,
                                },
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:0002",
                            "hasNextPage": False,
                        },
                    },
                }
            }
        }
    )
    assert events == [
        NewStarEvent(
            timestamp=datetime(2021, 11, 17, 15, 26, 39, tzinfo=timezone.utc),
            repo=repo,
            user=User(
                login="repo.fan",
                url="https://github.com/repo.fan",
                name="Fan of Repositories",
                isViewer=False,
            ),
        ),
        NewStarEvent(
            timestamp=datetime(2021, 11, 17, 23, 45, 56, tzinfo=timezone.utc),
            repo=repo,
            user=User(
                login="starry.user",
                url="https://github.com/starry.user",
                name=None,
                isViewer=False,
            ),
        ),
    ]
    assert manager.cursors == {ActivityType.STAR: "cursor:0002"}
    assert not manager.has_next_page


def test_activity_query_stargazers_no_cursor() -> None:
    q = (
        "query($repo_id: ID!) {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            stargazers(\n"
        "                orderBy: {field: STARRED_AT, direction: ASC},\n"
        "                last: 1\n"
        "            ) {\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(repo=repo, types=[ActivityType.STAR], cursors={})
    assert manager.has_next_page
    assert manager.make_query() == (q, {"repo_id": "id:viewer/repo"})


def test_activity_query_forks() -> None:
    q = (
        "query(\n"
        "    $repo_id: ID!,\n"
        "    $page_size: Int!,\n"
        "    $forks_cursor: String\n"
        ") {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            forks(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                first: $page_size,\n"
        "                after: $forks_cursor\n"
        "            ) {\n"
        "                nodes {\n"
        "                    createdAt\n"
        "                    id\n"
        "                    nameWithOwner\n"
        "                    owner {\n"
        "                        login\n"
        "                        url\n"
        "                        ... on User {\n"
        "                            name\n"
        "                            isViewer\n"
        "                        }\n"
        "                        ... on Organization {\n"
        "                            name\n"
        "                        }\n"
        "                    }\n"
        "                    name\n"
        "                    url\n"
        "                    description\n"
        "                    descriptionHTML\n"
        "                }\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                    hasNextPage\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(
        repo=repo,
        types=[ActivityType.FORK],
        cursors={ActivityType.FORK: "cursor:0001"},
    )
    assert manager.has_next_page
    assert manager.make_query() == (
        q,
        {
            "repo_id": "id:viewer/repo",
            "page_size": PAGE_SIZE,
            "forks_cursor": "cursor:0001",
        },
    )
    events = manager.parse_response(
        {
            "data": {
                "node": {
                    "forks": {
                        "nodes": [
                            {
                                "id": "id:archive/repo",
                                "createdAt": "2021-11-17T15:33:53Z",
                                "nameWithOwner": "archive/repo",
                                "owner": {
                                    "login": "archive",
                                    "url": "https://github.com/archive",
                                    "name": "The Repository Archive Org",
                                },
                                "name": "repo",
                                "url": "https://github.com/archive/repo",
                                "description": 'Archive of "My Very Special Repo(tm)"',
                                "descriptionHTML": (
                                    '<div>Archive of "My Very Special Repo(tm)"</div>'
                                ),
                            },
                            {
                                "id": "id:forker/repofork",
                                "createdAt": "2021-11-17T16:43:54Z",
                                "nameWithOwner": "forker/repofork",
                                "owner": {
                                    "login": "forker",
                                    "url": "https://github.com/forker",
                                    "name": "I Fork Repositories",
                                    "isViewer": False,
                                },
                                "name": "repofork",
                                "url": "https://github.com/archive/repofork",
                                "description": "My Very Special Repo(tm)",
                                "descriptionHTML": (
                                    "<div>My Very Special Repo(tm)</div>"
                                ),
                            },
                        ],
                        "pageInfo": {
                            "endCursor": "cursor:0002",
                            "hasNextPage": False,
                        },
                    },
                }
            }
        }
    )
    assert events == [
        NewForkEvent(
            timestamp=datetime(2021, 11, 17, 15, 33, 53, tzinfo=timezone.utc),
            repo=repo,
            fork=Repository(
                id="id:archive/repo",
                nameWithOwner="archive/repo",
                owner=User(
                    login="archive",
                    url="https://github.com/archive",
                    name="The Repository Archive Org",
                    isViewer=False,
                ),
                name="repo",
                url="https://github.com/archive/repo",
                description='Archive of "My Very Special Repo(tm)"',
                descriptionHTML='<div>Archive of "My Very Special Repo(tm)"</div>',
            ),
        ),
        NewForkEvent(
            timestamp=datetime(2021, 11, 17, 16, 43, 54, tzinfo=timezone.utc),
            repo=repo,
            fork=Repository(
                id="id:forker/repofork",
                nameWithOwner="forker/repofork",
                owner=User(
                    login="forker",
                    url="https://github.com/forker",
                    name="I Fork Repositories",
                    isViewer=False,
                ),
                name="repofork",
                url="https://github.com/archive/repofork",
                description="My Very Special Repo(tm)",
                descriptionHTML="<div>My Very Special Repo(tm)</div>",
            ),
        ),
    ]
    assert manager.cursors == {ActivityType.FORK: "cursor:0002"}
    assert not manager.has_next_page


def test_activity_query_forks_no_cursor() -> None:
    q = (
        "query($repo_id: ID!) {\n"
        "    node(id: $repo_id) {\n"
        "        ... on Repository {\n"
        "            forks(\n"
        "                orderBy: {field: CREATED_AT, direction: ASC},\n"
        "                last: 1\n"
        "            ) {\n"
        "                pageInfo {\n"
        "                    endCursor\n"
        "                }\n"
        "            }\n"
        "        }\n"
        "    }\n"
        "}\n"
    )
    repo = Repository(
        id="id:viewer/repo",
        nameWithOwner="viewer/repo",
        owner=User(
            login="viewer",
            url="https://github.com/viewer",
            name="Vid Ewer",
            isViewer=True,
        ),
        name="repo",
        url="https://github.com/viewer/repo",
        description="My Very Special Repo(tm)",
        descriptionHTML="<div>My Very Special Repo(tm)</div>",
    )
    manager = ActivityQuery(repo=repo, types=[ActivityType.FORK], cursors={})
    assert manager.has_next_page
    assert manager.make_query() == (q, {"repo_id": "id:viewer/repo"})



from __future__ import annotations
from collections.abc import Iterator
from email import message_from_string, policy
from email.message import EmailMessage
from functools import partial
import json
import logging
from pathlib import Path
from shutil import copytree
from traceback import format_exception
from types import TracebackType
from typing import Dict, List, Optional, Union, cast
from unittest.mock import ANY
from click.testing import CliRunner, Result
from mailbits import email2dict
from pydantic import BaseModel
import pytest
from pytest_mock import MockerFixture
from reponews.__main__ import main
from reponews.types import (
    ActivityType,
    Affiliation,
    CursorDict,
    NewDiscussionEvent,
    NewForkEvent,
    NewIssueEvent,
    NewPREvent,
    NewReleaseEvent,
    NewStarEvent,
    NewTagEvent,
    RepoActivity,
    Repository,
)
from reponews.util import MAIL_USER_AGENT, NotFoundError

MOCK_DIR = Path(__file__).with_name("data") / "mock"


class ActivityQuery(BaseModel):
    activity_types: List[ActivityType]
    cursors_in: CursorDict
    events: List[
        Union[
            NewDiscussionEvent,
            NewForkEvent,
            NewIssueEvent,
            NewPREvent,
            NewReleaseEvent,
            NewStarEvent,
            NewTagEvent,
        ]
    ]
    cursors_out: CursorDict


class SessionData(BaseModel):
    affiliated: List[Repository]
    owners: Dict[str, Optional[List[Repository]]]
    repos: Dict[str, Optional[Repository]]
    activity: Dict[str, ActivityQuery]


class MockClient:
    def __init__(self, api_url: str, token: str, querying: bool = True) -> None:
        assert api_url == "https://test.nil/api"
        assert token == "1234567890"
        with (MOCK_DIR / "session-data.json").open() as fp:
            self.data = SessionData.model_validate(json.load(fp))
        self.querying = querying

    def __enter__(self) -> MockClient:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        _exc_val: BaseException | None,
        _exc_tb: TracebackType | None,
    ) -> None:
        if exc_type is None and self.querying:
            assert not self.data.owners, "Not all owners queried"
            assert not self.data.repos, "Not all repositories queried"
            assert not self.data.activity, "Not all activity queried"

    def get_affiliated_repos(
        self, affiliations: list[Affiliation]
    ) -> Iterator[Repository]:
        assert affiliations == [Affiliation.OWNER, Affiliation.COLLABORATOR]
        return iter(self.data.affiliated)

    def get_owner_repos(self, owner: str) -> Iterator[Repository]:
        assert owner in self.data.owners
        repos = self.data.owners.pop(owner)
        if repos is None:
            raise NotFoundError(f"No such repository owner: {owner}")
        return iter(repos)

    def get_repo(self, owner: str, name: str) -> Repository:
        fullname = f"{owner}/{name}"
        assert fullname in self.data.repos
        r = self.data.repos.pop(fullname)
        if r is None:
            raise NotFoundError(f"No such repository: {fullname}")
        return r

    def get_new_repo_activity(
        self, repo: Repository, types: list[ActivityType], cursors: CursorDict
    ) -> tuple[list[RepoActivity], CursorDict]:
        fullname = repo.nameWithOwner
        assert fullname in self.data.activity
        aq = self.data.activity.pop(fullname)
        assert aq.activity_types == types
        assert aq.cursors_in == cursors
        return (cast(List[RepoActivity], aq.events), aq.cursors_out)


MSG_SPEC = {
    "unixfrom": None,
    "headers": {
        "subject": "Your Repo News is here!",
        "from": [
            {
                "display_name": "RepoNews",
                "address": "reponews@example.com",
            },
        ],
        "to": [
            {
                "display_name": "",
                "address": "viewer@example.org",
            }
        ],
        "content-type": {
            "content_type": "text/plain",
            "params": {},
        },
        "user-agent": [MAIL_USER_AGENT],
    },
    "preamble": ANY,
    "content": (MOCK_DIR / "body.txt").read_text(),
    "epilogue": ANY,
}


def show_result(r: Result) -> str:
    if r.exception is not None:
        assert isinstance(r.exc_info, tuple)
        return "".join(format_exception(*r.exc_info))
    else:
        return r.output


def assert_json_eq(p1: Path, p2: Path) -> None:
    assert json.loads(p1.read_text()) == json.loads(p2.read_text())


def test_dump_repos(
    caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr("reponews.client.Client", partial(MockClient, querying=False))
    copytree(MOCK_DIR / "home", tmp_path / "home")
    r = CliRunner().invoke(
        main,
        ["--config", str(tmp_path / "home" / "config.toml"), "--dump-repos"],
        standalone_mode=False,
    )
    assert r.exit_code == 0, show_result(r)
    assert r.output == (MOCK_DIR / "dump-repos.txt").read_text()
    assert (
        "reponews",
        logging.WARNING,
        "User user-not-found does not exist!",
    ) in caplog.record_tuples
    assert (
        "reponews",
        logging.WARNING,
        "Repository luser/repo-not-found does not exist!",
    ) in caplog.record_tuples
    assert_json_eq(tmp_path / "home" / "state.json", MOCK_DIR / "home" / "state.json")


@pytest.mark.parametrize("save", [True, False])
def test_print_body(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    save: bool,
) -> None:
    monkeypatch.setattr("reponews.client.Client", MockClient)
    copytree(MOCK_DIR / "home", tmp_path / "home")
    args = ["--config", str(tmp_path / "home" / "config.toml"), "--print-body"]
    if not save:
        args.append("--no-save")
    r = CliRunner().invoke(main, args, standalone_mode=False)
    assert r.exit_code == 0, show_result(r)
    assert r.output == (MOCK_DIR / "body.txt").read_text()
    assert (
        "reponews",
        logging.WARNING,
        "User user-not-found does not exist!",
    ) in caplog.record_tuples
    assert (
        "reponews",
        logging.WARNING,
        "Repository luser/repo-not-found does not exist!",
    ) in caplog.record_tuples
    if save:
        assert_json_eq(tmp_path / "home" / "state.json", MOCK_DIR / "new-state.json")
    else:
        assert_json_eq(
            tmp_path / "home" / "state.json", MOCK_DIR / "home" / "state.json"
        )


@pytest.mark.parametrize("save", [True, False])
def test_print(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    save: bool,
) -> None:
    monkeypatch.setattr("reponews.client.Client", MockClient)
    copytree(MOCK_DIR / "home", tmp_path / "home")
    args = ["--config", str(tmp_path / "home" / "config.toml"), "--print"]
    if not save:
        args.append("--no-save")
    r = CliRunner().invoke(main, args, standalone_mode=False)
    assert r.exit_code == 0, show_result(r)
    msg = message_from_string(r.output, policy=policy.default)
    assert email2dict(msg) == MSG_SPEC
    assert (
        "reponews",
        logging.WARNING,
        "User user-not-found does not exist!",
    ) in caplog.record_tuples
    assert (
        "reponews",
        logging.WARNING,
        "Repository luser/repo-not-found does not exist!",
    ) in caplog.record_tuples
    if save:
        assert_json_eq(tmp_path / "home" / "state.json", MOCK_DIR / "new-state.json")
    else:
        assert_json_eq(
            tmp_path / "home" / "state.json", MOCK_DIR / "home" / "state.json"
        )


@pytest.mark.parametrize("save", [True, False])
def test_send(
    caplog: pytest.LogCaptureFixture,
    mocker: MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    save: bool,
) -> None:
    monkeypatch.setattr("reponews.client.Client", MockClient)
    m = mocker.patch("outgoing.senders.null.NullSender", autospec=True)
    copytree(MOCK_DIR / "home", tmp_path / "home")
    args = ["--config", str(tmp_path / "home" / "config.toml")]
    if not save:
        args.append("--no-save")
    r = CliRunner().invoke(main, args, standalone_mode=False)
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    m.assert_called_once_with(
        method="null", configpath=tmp_path / "home" / "config.toml"
    )
    instance = m.return_value
    assert instance.send.call_count == 1
    sent = instance.send.call_args[0][0]
    assert isinstance(sent, EmailMessage)
    assert email2dict(sent) == MSG_SPEC
    assert (
        "reponews",
        logging.WARNING,
        "User user-not-found does not exist!",
    ) in caplog.record_tuples
    assert (
        "reponews",
        logging.WARNING,
        "Repository luser/repo-not-found does not exist!",
    ) in caplog.record_tuples
    if save:
        assert_json_eq(tmp_path / "home" / "state.json", MOCK_DIR / "new-state.json")
    else:
        assert_json_eq(
            tmp_path / "home" / "state.json", MOCK_DIR / "home" / "state.json"
        )



from __future__ import annotations
from textwrap import indent

INDENT = " " * 4
SINGLE_LINE_CUTOFF = 40


class Object:
    def __init__(self, name: str, args: dict[str, str], *fields: str | Object) -> None:
        self.name = name
        self.args = args
        self.fields = fields

    def __str__(self) -> str:
        s = self.name
        if self.args:
            argslist = [f"{k}: {v}" for k, v in self.args.items()]
            if len(", ".join(argslist)) > SINGLE_LINE_CUTOFF:
                s += "(\n" + ",\n".join(INDENT + a for a in argslist) + "\n)"
            else:
                s += "(" + ", ".join(argslist) + ")"
        s += " {\n"
        for f in self.fields:
            s += indent(str(f).rstrip(), INDENT) + "\n"
        s += "}\n"
        return s


def mkissueoid_connection(name: str) -> Object:
    return Object(
        name,
        {
            "orderBy": "{field: CREATED_AT, direction: ASC}",
            "first": "$page_size",
            "after": f"${name}_cursor",
        },
        Object(
            "nodes",
            {},
            Object("author", {}, *ACTOR_FIELDS),
            "createdAt",
            "number",
            "title",
            "url",
        ),
        Object("pageInfo", {}, "endCursor", "hasNextPage"),
    )


def mklastconn(name: str) -> Object:
    return Object(
        name,
        {
            "orderBy": "{field: CREATED_AT, direction: ASC}",
            "last": "1",
        },
        Object("pageInfo", {}, "endCursor"),
    )


ACTOR_FIELDS: list[str | Object] = [
    "login",
    "url",
    Object("... on User", {}, "name", "isViewer"),
]

OWNER_FIELDS: list[str | Object] = [
    "login",
    "url",
    Object("... on User", {}, "name", "isViewer"),
    Object("... on Organization", {}, "name"),
]

USER_FIELDS = ["login", "url", "name", "isViewer"]

REPO_FIELDS: list[str | Object] = [
    "id",
    "nameWithOwner",
    Object("owner", {}, *OWNER_FIELDS),
    "name",
    "url",
    "description",
    "descriptionHTML",
]

ISSUE_CONNECTION = mkissueoid_connection("issues")

ISSUE_LAST_CONNECTION = mklastconn("issues")

PR_CONNECTION = mkissueoid_connection("pullRequests")

PR_LAST_CONNECTION = mklastconn("pullRequests")

DISCUSSION_CONNECTION = mkissueoid_connection("discussions")

DISCUSSION_LAST_CONNECTION = mklastconn("discussions")

RELEASE_CONNECTION = Object(
    "releases",
    {
        "orderBy": "{field: CREATED_AT, direction: ASC}",
        "first": "$page_size",
        "after": "$releases_cursor",
    },
    Object(
        "nodes",
        {},
        "name",
        "tagName",
        Object("author", {}, *USER_FIELDS),
        "createdAt",
        "description",
        "descriptionHTML",
        "isPrerelease",
        "isDraft",
        "url",
    ),
    Object("pageInfo", {}, "endCursor", "hasNextPage"),
)

RELEASE_LAST_CONNECTION = mklastconn("releases")

TAG_CONNECTION = Object(
    "tags: refs",
    {
        "refPrefix": '"refs/tags/"',
        "orderBy": "{field: TAG_COMMIT_DATE, direction: ASC}",
        "first": "$page_size",
        "after": "$tags_cursor",
    },
    Object(
        "nodes",
        {},
        "name",
        Object(
            "target",
            {},
            "__typename",
            Object(
                "... on Tag",
                {},
                Object(
                    "tagger",
                    {},
                    "date",
                    Object("user", {}, *USER_FIELDS),
                ),
            ),
            Object(
                "... on Commit",
                {},
                "committedDate",
                Object("author", {}, Object("user", {}, *USER_FIELDS)),
            ),
        ),
    ),
    Object("pageInfo", {}, "endCursor", "hasNextPage"),
)

TAG_LAST_CONNECTION = Object(
    "tags: refs",
    {
        "refPrefix": '"refs/tags/"',
        "orderBy": "{field: TAG_COMMIT_DATE, direction: ASC}",
        "last": "1",
    },
    Object("pageInfo", {}, "endCursor"),
)

STAR_CONNECTION = Object(
    "stargazers",
    {
        "orderBy": "{field: STARRED_AT, direction: ASC}",
        "first": "$page_size",
        "after": "$stargazers_cursor",
    },
    Object("edges", {}, "starredAt", Object("user: node", {}, *USER_FIELDS)),
    Object("pageInfo", {}, "endCursor", "hasNextPage"),
)

STAR_LAST_CONNECTION = Object(
    "stargazers",
    {
        "orderBy": "{field: STARRED_AT, direction: ASC}",
        "last": "1",
    },
    Object("pageInfo", {}, "endCursor"),
)

FORK_CONNECTION = Object(
    "forks",
    {
        "orderBy": "{field: CREATED_AT, direction: ASC}",
        "first": "$page_size",
        "after": "$forks_cursor",
    },
    Object("nodes", {}, "createdAt", *REPO_FIELDS),
    Object("pageInfo", {}, "endCursor", "hasNextPage"),
)

FORK_LAST_CONNECTION = mklastconn("forks")



from __future__ import annotations
import json
from operator import attrgetter
from os.path import expanduser
from pathlib import Path
from typing import List, Tuple
import ghtoken
from pydantic import BaseModel, ValidationError
import pytest
from pytest_mock import MockerFixture
from reponews.config import ActivityPrefs, Configuration
from reponews.types import ActivityType, Repository, User
from reponews.util import UserError, get_default_state_file

DATA_DIR = Path(__file__).with_name("data")


def mkrepo(owner: str, name: str) -> Repository:
    return Repository(
        id="1234",
        owner=User(login=owner, url=f"https://github.com/{owner}"),
        name=name,
        nameWithOwner=f"{owner}/{name}",
        url=f"https://github.com/{owner}/{name}",
        description="A repository",
        descriptionHTML="A repository",
    )


@pytest.mark.parametrize(
    "tomlfile",
    sorted((DATA_DIR / "config").glob("*.toml")),
    ids=attrgetter("name"),
)
@pytest.mark.usefixtures("tmp_home")
def test_parse_config(tomlfile: Path) -> None:
    config = Configuration.from_toml_file(tomlfile)
    with tomlfile.with_suffix(".json").open() as fp:
        expected = json.load(fp)
    if expected["state_file"] is None:
        expected["state_file"] = get_default_state_file()
    for key in ("auth_token_file", "state_file"):
        if expected[key] is not None:
            expected[key] = expanduser(expected[key]).format(config=tomlfile.parent)
    assert config.model_dump(mode="json") == expected


@pytest.mark.parametrize(
    "tomlfile",
    sorted((DATA_DIR / "bad-config").glob("*.toml")),
    ids=attrgetter("name"),
)
def test_parse_bad_config(tomlfile: Path) -> None:
    with pytest.raises(ValidationError):
        Configuration.from_toml_file(tomlfile)


class InclusionCase(BaseModel):
    included_owners: List[str]
    included_repos: List[Tuple[str, str]]
    not_excluded_repos: List[Tuple[str, str]]
    excluded_repos: List[Tuple[str, str]]


@pytest.mark.parametrize(
    "tomlfile",
    sorted((DATA_DIR / "inclusions").glob("*.toml")),
    ids=attrgetter("name"),
)
def test_inclusions(tomlfile: Path) -> None:
    config = Configuration.from_toml_file(DATA_DIR / "inclusions" / tomlfile)
    with tomlfile.with_suffix(".json").open() as fp:
        expected = InclusionCase.model_validate(json.load(fp))
    assert config.get_included_repo_owners() == expected.included_owners
    assert config.get_included_repos() == expected.included_repos
    for owner, name in expected.not_excluded_repos:
        assert not config.is_repo_excluded(mkrepo(owner, name))
    for owner, name in expected.excluded_repos:
        assert config.is_repo_excluded(mkrepo(owner, name))


@pytest.mark.parametrize(
    "tomlfile,repo,is_affiliated,prefs",
    [
        (
            "empty.toml",
            mkrepo("owner", "repo"),
            True,
            ActivityPrefs(
                issues=True,
                pull_requests=True,
                discussions=True,
                releases=True,
                tags=True,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=False,
            ),
        ),
        (
            "empty.toml",
            mkrepo("owner", "repo"),
            False,
            ActivityPrefs(
                issues=True,
                pull_requests=True,
                discussions=True,
                releases=True,
                tags=True,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=False,
            ),
        ),
        (
            "full.toml",
            mkrepo("owner", "repo"),
            True,
            ActivityPrefs(
                issues=False,
                pull_requests=False,
                discussions=True,
                releases=True,
                tags=True,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=True,
            ),
        ),
        (
            "full.toml",
            mkrepo("owner", "repo"),
            False,
            ActivityPrefs(
                issues=False,
                pull_requests=False,
                discussions=True,
                releases=True,
                tags=True,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=True,
            ),
        ),
        (
            "repo-activity01.toml",
            mkrepo("luser", "repo"),
            True,
            ActivityPrefs(
                issues=False,
                pull_requests=False,
                discussions=True,
                releases=True,
                tags=True,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=False,
            ),
        ),
        (
            "repo-activity01.toml",
            mkrepo("luser", "repo"),
            False,
            ActivityPrefs(
                issues=True,
                pull_requests=True,
                discussions=True,
                releases=True,
                tags=True,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=False,
            ),
        ),
        (
            "repo-activity01.toml",
            mkrepo("owner", "project"),
            False,
            ActivityPrefs(
                issues=True,
                pull_requests=True,
                discussions=True,
                releases=True,
                tags=False,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=True,
            ),
        ),
        (
            "repo-activity01.toml",
            mkrepo("owner", "repo"),
            False,
            ActivityPrefs(
                issues=True,
                pull_requests=True,
                discussions=True,
                releases=False,
                tags=False,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=False,
            ),
        ),
        (
            "repo-activity01.toml",
            mkrepo("owner", "repo"),
            True,
            ActivityPrefs(
                issues=False,
                pull_requests=False,
                discussions=True,
                releases=False,
                tags=False,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=False,
            ),
        ),
        (
            "repo-activity02.toml",
            mkrepo("luser", "repo"),
            True,
            ActivityPrefs(
                issues=True,
                pull_requests=False,
                discussions=False,
                releases=True,
                tags=True,
                released_tags=False,
                stars=False,
                forks=False,
                my_activity=False,
            ),
        ),
        (
            "repo-activity02.toml",
            mkrepo("luser", "repo"),
            False,
            ActivityPrefs(
                issues=False,
                pull_requests=True,
                discussions=False,
                releases=True,
                tags=True,
                released_tags=False,
                stars=False,
                forks=False,
                my_activity=False,
            ),
        ),
        (
            "repo-activity02.toml",
            mkrepo("owner", "project"),
            False,
            ActivityPrefs(
                issues=False,
                pull_requests=True,
                discussions=False,
                releases=True,
                tags=False,
                released_tags=False,
                stars=True,
                forks=False,
                my_activity=True,
            ),
        ),
        (
            "repo-activity02.toml",
            mkrepo("owner", "repo"),
            False,
            ActivityPrefs(
                issues=False,
                pull_requests=True,
                discussions=False,
                releases=False,
                tags=False,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=False,
            ),
        ),
        (
            "repo-activity02.toml",
            mkrepo("owner", "repo"),
            True,
            ActivityPrefs(
                issues=True,
                pull_requests=False,
                discussions=False,
                releases=False,
                tags=False,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=False,
            ),
        ),
        (
            "yestags.toml",
            mkrepo("owner", "repo"),
            True,
            ActivityPrefs(
                issues=True,
                pull_requests=True,
                discussions=True,
                releases=True,
                tags=True,
                released_tags=True,
                stars=True,
                forks=True,
                my_activity=False,
            ),
        ),
        (
            "yestags.toml",
            mkrepo("owner", "repo"),
            False,
            ActivityPrefs(
                issues=True,
                pull_requests=True,
                discussions=True,
                releases=True,
                tags=True,
                released_tags=True,
                stars=True,
                forks=True,
                my_activity=False,
            ),
        ),
    ],
)
def test_get_repo_activity_prefs(
    tomlfile: str, repo: Repository, is_affiliated: bool, prefs: ActivityPrefs
) -> None:
    config = Configuration.from_toml_file(DATA_DIR / "config" / tomlfile)
    assert config.get_repo_activity_prefs(repo, is_affiliated) == prefs


@pytest.mark.parametrize(
    "prefs,types",
    [
        (
            ActivityPrefs(
                issues=True,
                pull_requests=True,
                discussions=True,
                releases=True,
                tags=True,
                released_tags=False,
                stars=True,
                forks=True,
                my_activity=False,
            ),
            [
                ActivityType.ISSUE,
                ActivityType.PR,
                ActivityType.DISCUSSION,
                ActivityType.RELEASE,
                ActivityType.TAG,
                ActivityType.STAR,
                ActivityType.FORK,
            ],
        ),
        (
            ActivityPrefs(
                issues=False,
                pull_requests=False,
                discussions=False,
                releases=False,
                tags=False,
                released_tags=True,
                stars=False,
                forks=False,
                my_activity=True,
            ),
            [],
        ),
        (
            ActivityPrefs(
                issues=True,
                pull_requests=True,
                discussions=True,
                releases=True,
                tags=False,
                released_tags=True,
                stars=False,
                forks=False,
                my_activity=True,
            ),
            [
                ActivityType.ISSUE,
                ActivityType.PR,
                ActivityType.DISCUSSION,
                ActivityType.RELEASE,
            ],
        ),
    ],
)
def test_get_activity_types(prefs: ActivityPrefs, types: list[ActivityType]) -> None:
    assert prefs.get_activity_types() == types


def test_get_auth_token_explicit(
    monkeypatch: pytest.MonkeyPatch, tmp_home: Path
) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "QWERTY")
    (tmp_home / "token.txt").write_text("abcdef\n")
    config = Configuration(
        auth_token="123456",
        auth_token_file="~/token.txt",
    )
    assert config.get_auth_token() == "123456"


def test_get_auth_token_file(monkeypatch: pytest.MonkeyPatch, tmp_home: Path) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "QWERTY")
    (tmp_home / "token.txt").write_text("abcdef\n")
    config = Configuration(auth_token_file="~/token.txt")
    assert config.get_auth_token() == "abcdef"


@pytest.mark.usefixtures("tmp_home")
def test_get_auth_token_missing_file(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "QWERTY")
    config = Configuration(auth_token_file="~/token.txt")
    with pytest.raises(FileNotFoundError):
        config.get_auth_token()


def test_get_auth_token_ghtoken(mocker: MockerFixture, tmp_home: Path) -> None:
    m = mocker.patch("ghtoken.get_ghtoken", return_value="gh_token")
    (tmp_home / "token.txt").write_text("abcdef\n")
    config = Configuration()
    assert config.get_auth_token() == "gh_token"
    m.assert_called_once_with(dotenv=False)


def test_get_auth_token_notset(mocker: MockerFixture, tmp_home: Path) -> None:
    m = mocker.patch("ghtoken.get_ghtoken", side_effect=ghtoken.GHTokenNotFound())
    (tmp_home / "token.txt").write_text("abcdef\n")
    config = Configuration()
    with pytest.raises(UserError) as excinfo:
        config.get_auth_token()
    assert str(excinfo.value) == (
        "GitHub access token not found.  Set via config file, GH_TOKEN,"
        " GITHUB_TOKEN, gh, hub, or hub.oauthtoken."
    )
    m.assert_called_once_with(dotenv=False)



from __future__ import annotations
from collections import defaultdict
from dataclasses import asdict, dataclass, field
from email.headerregistry import Address as PyAddress
from functools import cached_property
from pathlib import Path
import re
import sys
from typing import Any, Dict, List, Optional
from ghrepo import GH_REPO_RGX, GH_USER_RGX
import ghtoken  # Module import for mocking purposes
from mailbits import parse_address
from pydantic import AnyHttpUrl, BaseModel, Field, GetCoreSchemaHandler, field_validator
from pydantic.functional_serializers import PlainSerializer
from pydantic_core import CoreSchema, core_schema
from .types import ActivityType, Affiliation, Repository
from .util import UserError, default_api_url, get_default_state_file, mkalias

if sys.version_info[:2] >= (3, 11):
    from tomllib import load as toml_load
else:
    from tomli import load as toml_load

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated


@dataclass
class Address:
    name: Optional[str]
    address: str

    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls._parse,
            handler(str),
            serialization=core_schema.plain_serializer_function_ser_schema(asdict),
        )

    @classmethod
    def _parse(cls, value: str) -> Address:
        addr = parse_address(value)
        return cls(name=addr.display_name or None, address=addr.addr_spec)

    def as_py_address(self) -> PyAddress:
        return PyAddress(self.name or "", addr_spec=self.address)


@dataclass(frozen=True)
class RepoSpec:
    owner: str
    name: Optional[str]

    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls.parse,
            handler(str),
            serialization=core_schema.plain_serializer_function_ser_schema(asdict),
        )

    @classmethod
    def parse(cls, value: str) -> RepoSpec:
        m = re.fullmatch(
            rf"(?P<owner>{GH_USER_RGX})/(?:\*|(?P<name>{GH_REPO_RGX}))", value
        )
        if m:
            return cls(owner=m["owner"], name=m["name"])
        else:
            raise ValueError(f"Invalid repo spec: {value!r}")

    def __str__(self) -> str:
        return f"{self.owner}/{self.name or '*'}"


class BaseConfig(BaseModel):
    model_config = {
        "alias_generator": mkalias,
        "populate_by_name": True,
        "extra": "forbid",
    }


class PartialActivityPrefs(BaseConfig):
    issues: Optional[bool] = None
    pull_requests: Optional[bool] = None
    discussions: Optional[bool] = None
    releases: Optional[bool] = None
    tags: Optional[bool] = None
    released_tags: Optional[bool] = None
    stars: Optional[bool] = None
    forks: Optional[bool] = None
    my_activity: Optional[bool] = None


class ActivityPrefs(PartialActivityPrefs):
    issues: bool = True
    pull_requests: bool = True
    discussions: bool = True
    releases: bool = True
    tags: bool = True
    released_tags: bool = False
    stars: bool = True
    forks: bool = True
    my_activity: bool = False

    def get_activity_types(self) -> list[ActivityType]:
        types: list[ActivityType] = []
        if self.issues:
            types.append(ActivityType.ISSUE)
        if self.pull_requests:
            types.append(ActivityType.PR)
        if self.discussions:
            types.append(ActivityType.DISCUSSION)
        if self.releases:
            types.append(ActivityType.RELEASE)
        if self.tags:
            types.append(ActivityType.TAG)
        if self.stars:
            types.append(ActivityType.STAR)
        if self.forks:
            types.append(ActivityType.FORK)
        return types

    def update(self, prefs: PartialActivityPrefs) -> None:
        pd = dict(prefs)
        for field_name in self.model_fields:
            if (v := pd[field_name]) is not None:
                assert isinstance(v, bool)
                setattr(self, field_name, v)


class RepoActivityPrefs(PartialActivityPrefs):
    include: bool = True


class ActivityConfig(PartialActivityPrefs):
    affiliated: PartialActivityPrefs = Field(default_factory=PartialActivityPrefs)
    repo: Dict[
        Annotated[RepoSpec, PlainSerializer(lambda s: str(s))], RepoActivityPrefs
    ] = Field(default_factory=dict)


class ReposConfig(BaseConfig):
    affiliations: List[Affiliation] = Field(default_factory=lambda: list(Affiliation))
    include: List[RepoSpec] = Field(default_factory=list)
    exclude: List[RepoSpec] = Field(default_factory=list)


@dataclass
class RepoInclusions:
    included_owners: set[str] = field(default_factory=set)
    excluded_owners: set[str] = field(default_factory=set)
    included_repos: dict[str, set[str]] = field(
        default_factory=lambda: defaultdict(set)
    )
    excluded_repos: dict[str, set[str]] = field(
        default_factory=lambda: defaultdict(set)
    )

    @classmethod
    def from_repos_config(
        cls, repos_config: ReposConfig, preffed_repos: list[RepoSpec]
    ) -> RepoInclusions:
        rinc = cls()
        for rs in repos_config.include + [
            RepoSpec(owner=r.owner, name=r.name) for r in preffed_repos
        ]:
            if rs.name is None:
                rinc.included_owners.add(rs.owner)
            else:
                rinc.included_repos[rs.owner].add(rs.name)
        for rs in repos_config.exclude:
            if rs.name is None:
                rinc.excluded_owners.add(rs.owner)
                rinc.included_owners.discard(rs.owner)
                rinc.included_repos.pop(rs.owner, None)
            else:
                rinc.excluded_repos[rs.owner].add(rs.name)
                rinc.included_repos[rs.owner].discard(rs.name)
        return rinc

    def get_included_repo_owners(self) -> list[str]:
        return sorted(self.included_owners)

    def get_included_repos(self) -> list[tuple[str, str]]:
        return [
            (owner, n)
            for owner, names in sorted(self.included_repos.items())
            if owner not in self.included_owners
            for n in sorted(names)
        ]

    def is_repo_excluded(self, repo: Repository) -> bool:
        return (
            repo.owner.login in self.excluded_owners
            or repo.name in self.excluded_repos[repo.owner.login]
        )


class Configuration(BaseConfig):
    recipient: Optional[Address] = None
    sender: Optional[Address] = None
    subject: str = "[reponews] New activity on your GitHub repositories"
    auth_token: Optional[str] = None
    auth_token_file: Optional[Path] = None
    # The default is implemented as a factory in order to make it easy to test
    # with a fake $HOME:
    state_file: Path = Field(default_factory=get_default_state_file)
    api_url: AnyHttpUrl = Field(default_factory=default_api_url)
    activity: ActivityConfig = Field(default_factory=ActivityConfig)
    repos: ReposConfig = Field(default_factory=ReposConfig)

    @field_validator("auth_token_file", "state_file")
    @classmethod
    def _expand_path(cls, v: Optional[Path]) -> Optional[Path]:
        return v.expanduser() if v is not None else v

    @classmethod
    def from_toml_file(cls, filepath: str | Path) -> Configuration:
        with open(filepath, "rb") as fp:
            data = toml_load(fp).get("reponews", {})
        basedir = Path(filepath).parent
        config = cls.model_validate(data)
        if config.auth_token_file is not None:
            config.auth_token_file = basedir / config.auth_token_file
        config.state_file = basedir / config.state_file
        return config

    def get_auth_token(self) -> str:
        if self.auth_token is not None:
            return self.auth_token
        elif self.auth_token_file is not None:
            return self.auth_token_file.read_text(encoding="utf-8").strip()
        else:
            try:
                return ghtoken.get_ghtoken(dotenv=False)
            except ghtoken.GHTokenNotFound:
                raise UserError(
                    "GitHub access token not found.  Set via config file,"
                    " GH_TOKEN, GITHUB_TOKEN, gh, hub, or hub.oauthtoken."
                )

    def get_repo_activity_prefs(
        self, repo: Repository, is_affiliated: bool
    ) -> ActivityPrefs:
        prefs = ActivityPrefs()
        prefs.update(self.activity)
        if is_affiliated:
            prefs.update(self.activity.affiliated)
        for spec in [f"{repo.owner}/*", str(repo)]:
            try:
                p = self.activity.repo[RepoSpec.parse(spec)]
            except KeyError:
                pass
            else:
                prefs.update(p)
        return prefs

    @cached_property
    def inclusions(self) -> RepoInclusions:
        return RepoInclusions.from_repos_config(
            self.repos, [r for r, pref in self.activity.repo.items() if pref.include]
        )

    def get_included_repo_owners(self) -> list[str]:
        return self.inclusions.get_included_repo_owners()

    def get_included_repos(self) -> list[tuple[str, str]]:
        return self.inclusions.get_included_repos()

    def is_repo_excluded(self, repo: Repository) -> bool:
        return self.inclusions.is_repo_excluded(repo)



"""
Send e-mails about new events on your GitHub repositories

Do you want to receive e-mail notifications about new issues, pull requests,
discussions, releases, tags, stargazers, & forks on your GitHub repositories?
Of course you do — but turning on e-mail notifications in GitHub for
repositories you're watching will mean you get an e-mail for every comment on
every issue, which is a bit too much.  ``reponews`` aims for a happy medium:
only e-mailing you about new issues and similar activity — not about comments —
and only on repositories of your choice.  Simply set it up to run under cron or
another job scheduler (sold separately), point it at a compatible e-mail
sending service (also sold separately), and you'll get periodic e-mails listing
new events.

Visit <https://github.com/jwodder/reponews> for more information.
"""

__version__ = "0.4.0"
__author__ = "John Thorvald Wodder II"
__author_email__ = "reponews@varonathe.org"
__license__ = "MIT"
__url__ = "https://github.com/jwodder/reponews"



from __future__ import annotations
from enum import Enum
import json
import logging
from pathlib import Path
from typing import Optional
import click
from click_loglevel import LogLevel
from dotenv import find_dotenv, load_dotenv
from outgoing import Sender, from_config_file
from platformdirs import user_config_path
from . import __version__
from .core import RepoNews
from .util import UserError, log

DEFAULT_CONFIG_FILE = user_config_path("reponews", "jwodder") / "config.toml"

Mode = Enum("Mode", "PRINT PRINT_BODY DUMP_REPOS")


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(
    __version__,
    "-V",
    "--version",
    message="%(prog)s %(version)s",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    default=DEFAULT_CONFIG_FILE,
    show_default=True,
    help="Path to configuration file",
)
@click.option(
    "--dump-repos",
    "mode",
    flag_value=Mode.DUMP_REPOS,
    type=click.UNPROCESSED,
    help="List tracked repos and their activity preferences",
)
@click.option(
    "-E",
    "--env",
    type=click.Path(exists=True, dir_okay=False),
    help="Load environment variables from given .env file",
)
@click.option(
    "-l",
    "--log-level",
    type=LogLevel(),
    default="WARNING",
    help="Set logging level",
    show_default=True,
)
@click.option(
    "--print",
    "mode",
    flag_value=Mode.PRINT,
    type=click.UNPROCESSED,
    help="Output e-mail instead of sending",
)
@click.option(
    "--print-body",
    "mode",
    flag_value=Mode.PRINT_BODY,
    type=click.UNPROCESSED,
    help="Output e-mail body instead of sending",
)
@click.option(
    "--save/--no-save",
    default=True,
    help="Whether to update the state file  [default: --save]",
)
def main(
    config: Path, log_level: int, mode: Optional[Mode], save: bool, env: Optional[str]
) -> None:
    """
    Send e-mails about new events on your GitHub repositories.

    Visit <https://github.com/jwodder/reponews> for more information.
    """
    if env is None:
        env = find_dotenv(usecwd=True)
    load_dotenv(env)
    logging.basicConfig(
        format="%(asctime)s [%(levelname)-8s] %(name)s: %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
        level=log_level,
    )
    try:
        with RepoNews.from_config_file(config) as reponews:
            if mode is Mode.DUMP_REPOS:
                print(json.dumps(reponews.dump_repo_prefs(), indent=4, sort_keys=True))
                return
            if mode in (None, Mode.PRINT) and reponews.config.recipient is None:
                raise click.UsageError(
                    "reponews.recipient must be set when constructing an e-mail"
                )
            sender: Optional[Sender]
            if mode is None:
                # Fail early if the outgoing config is invalid or missing:
                sender = from_config_file(config, fallback=True)
                assert sender is not None
            else:
                sender = None
            events = reponews.get_new_activity()
            if events:
                if mode is Mode.PRINT:
                    print(reponews.compose_email(events))
                elif mode is Mode.PRINT_BODY:
                    print(reponews.compose_email_body(events))
                else:
                    log.info("Sending e-mail ...")
                    msg = reponews.compose_email(events)
                    assert sender is not None
                    with sender:
                        sender.send(msg)
            else:
                log.info("No new activity")
            if save:
                reponews.save_state()
    except UserError as e:
        raise click.UsageError(str(e))


if __name__ == "__main__":
    main()  # pragma: no cover



from __future__ import annotations
from collections.abc import Iterator
import json
from time import sleep
from types import TracebackType
from typing import Any, Optional
import requests
from .qmanager import (
    ActivityQuery,
    OwnersReposQuery,
    QueryManager,
    SingleRepoQuery,
    ViewersReposQuery,
)
from .types import ActivityType, Affiliation, CursorDict, RepoActivity, Repository
from .util import HTTP_USER_AGENT, NotFoundError, T, log

PAGE_SIZE = 50

MAX_RETRIES = 5
RETRY_STATUSES = (500, 502, 503, 504)
BACKOFF_FACTOR = 1.25
MAX_BACKOFF = 120


class Client:
    def __init__(self, api_url: str, token: str) -> None:
        self.api_url = api_url
        self.s = requests.Session()
        self.s.headers["Authorization"] = f"bearer {token}"
        self.s.headers["User-Agent"] = HTTP_USER_AGENT
        # <https://github.blog/2021-11-16-graphql-global-id-migration-update/>
        self.s.headers["X-Github-Next-Global-ID"] = "1"

    def __enter__(self) -> Client:
        return self

    def __exit__(
        self,
        _exc_type: type[BaseException] | None,
        _exc_val: BaseException | None,
        _exc_tb: TracebackType | None,
    ) -> None:
        self.s.close()

    def query(self, query: str, variables: Optional[dict[str, Any]] = None) -> Any:
        i = 0
        while True:
            r = self.s.post(
                self.api_url,
                json={"query": query, "variables": variables or {}},
            )
            if r.status_code in RETRY_STATUSES:
                if i + 1 < MAX_RETRIES:
                    delay = min(BACKOFF_FACTOR * 2**i, MAX_BACKOFF)
                    log.warning(
                        "GraphQL request returned %d; waiting %f seconds and retrying",
                        r.status_code,
                        delay,
                    )
                    sleep(delay)
                    i += 1
                    continue
                else:
                    log.error(
                        "GraphQL request returned %d; out of retries", r.status_code
                    )
                    raise APIException(r)
            elif not r.ok:
                raise APIException(r)
            else:
                break
        data = r.json()
        if data.get("errors"):
            try:
                # TODO: Figure out how to handle multi-errors where some of the
                # errors are NOT_FOUND
                if data["errors"][0]["type"] == "NOT_FOUND":
                    raise NotFoundError(data["errors"][0]["message"])
            except (AttributeError, LookupError, TypeError, ValueError):
                pass
            raise APIException(r)
        return data

    def do_managed_query(self, manager: QueryManager[T]) -> Iterator[T]:
        while manager.has_next_page:
            q, variables = manager.make_query()
            data = self.query(q, variables)
            yield from manager.parse_response(data)

    def get_affiliated_repos(
        self, affiliations: list[Affiliation]
    ) -> Iterator[Repository]:
        if not affiliations:
            log.info("No affiliations set; not fetching any affiliated repositories")
            return
        log.info(
            "Fetching repositories with affiliations %s",
            ", ".join(aff.value for aff in affiliations),
        )
        manager = ViewersReposQuery(affiliations=affiliations)
        for repo in self.do_managed_query(manager):
            log.info("Found repository %s", repo)
            yield repo

    def get_owner_repos(self, owner: str) -> Iterator[Repository]:
        log.info("Fetching repositories belonging to %s", owner)
        manager = OwnersReposQuery(owner=owner)
        for repo in self.do_managed_query(manager):
            log.info("Found repository %s", repo)
            yield repo

    def get_repo(self, owner: str, name: str) -> Repository:
        log.info("Fetching info for repo %s/%s", owner, name)
        manager = SingleRepoQuery(owner=owner, name=name)
        return next(self.do_managed_query(manager))

    def get_new_repo_activity(
        self, repo: Repository, types: list[ActivityType], cursors: CursorDict
    ) -> tuple[list[RepoActivity], CursorDict]:
        log.info(
            "Fetching new %s activity for %s", ", ".join(it.value for it in types), repo
        )
        manager = ActivityQuery(repo=repo, types=types, cursors=cursors)
        events: list[RepoActivity] = []
        for ev in self.do_managed_query(manager):
            log.info("Found activity on %s: %s", repo, ev)
            events.append(ev)
        return (events, manager.cursors)


class APIException(Exception):
    def __init__(self, response: requests.Response):
        self.response = response

    def __str__(self) -> str:
        if self.response.ok:
            msg = "GraphQL API error for URL: {0.url}\n"
        elif 400 <= self.response.status_code < 500:
            msg = "{0.status_code} Client Error: {0.reason} for URL: {0.url}\n"
        elif 500 <= self.response.status_code < 600:
            msg = "{0.status_code} Server Error: {0.reason} for URL: {0.url}\n"
        else:
            msg = "{0.status_code} Unknown Error: {0.reason} for URL: {0.url}\n"
        msg = msg.format(self.response)
        try:
            resp = self.response.json()
        except ValueError:
            msg += self.response.text
        else:
            msg += json.dumps(resp, sort_keys=True, indent=4)
        return msg



import logging
from pathlib import Path
import pytest


@pytest.fixture(autouse=True)
def capture_all_logs(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.DEBUG, logger="reponews")


@pytest.fixture()
def tmp_home(
    monkeypatch: pytest.MonkeyPatch, tmp_path_factory: pytest.TempPathFactory
) -> Path:
    home = tmp_path_factory.mktemp("tmp_home")
    monkeypatch.setenv("HOME", str(home))
    monkeypatch.setenv("USERPROFILE", str(home))
    monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
    monkeypatch.delenv("XDG_STATE_HOME", raising=False)
    monkeypatch.setenv("LOCALAPPDATA", str(home))
    return home



from __future__ import annotations
from abc import abstractmethod
from datetime import datetime
from enum import Enum
import json
import sys
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Literal, Optional
from eletter import reply_quote
from pydantic import BaseModel, field_validator
from pydantic.functional_serializers import PlainSerializer
from .qlobjs import (
    DISCUSSION_CONNECTION,
    DISCUSSION_LAST_CONNECTION,
    FORK_CONNECTION,
    FORK_LAST_CONNECTION,
    ISSUE_CONNECTION,
    ISSUE_LAST_CONNECTION,
    PR_CONNECTION,
    PR_LAST_CONNECTION,
    RELEASE_CONNECTION,
    RELEASE_LAST_CONNECTION,
    STAR_CONNECTION,
    STAR_LAST_CONNECTION,
    TAG_CONNECTION,
    TAG_LAST_CONNECTION,
    Object,
)
from .util import BogusEventError, dos2unix, log

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated


if TYPE_CHECKING:
    from typing_extensions import Self


class Affiliation(Enum):
    OWNER = "OWNER"
    ORGANIZATION_MEMBER = "ORGANIZATION_MEMBER"
    COLLABORATOR = "COLLABORATOR"


class User(BaseModel):
    # `name` is None/unset when the user is a bot or they never set their
    # display name
    name: Optional[str] = None
    login: str
    url: str
    isViewer: bool = False

    def __str__(self) -> str:
        return self.login


class Repository(BaseModel):
    id: str  # noqa: A003
    owner: User
    name: str
    nameWithOwner: str
    url: str
    description: Optional[str] = None
    descriptionHTML: str

    @classmethod
    def from_node(cls, node: dict[str, Any]) -> Repository:
        log.debug("Constructing Repository from node: %s", json.dumps(node))
        return cls.model_validate(node)

    def __str__(self) -> str:
        return self.nameWithOwner


class Event(BaseModel):
    event_type: str
    timestamp: datetime  # Used for sorting
    repo: Repository

    @abstractmethod
    def render(self) -> str:
        ...


class RepoActivity(Event):
    CONNECTION: ClassVar[Object]
    LAST_CONNECTION: ClassVar[Object]

    @classmethod
    @abstractmethod
    def from_node(cls, repo: Repository, node: dict[str, Any]) -> Self:
        ...

    @property
    @abstractmethod
    def is_mine(self) -> bool:
        ...


class NewIssueoidEvent(RepoActivity):
    number: int
    title: str
    author: User
    url: str

    @classmethod
    def from_node(cls, repo: Repository, node: dict[str, Any]) -> NewIssueoidEvent:
        log.debug("Constructing %s from node: %s", cls.__name__, json.dumps(node))
        node["timestamp"] = node.pop("createdAt")
        node["repo"] = repo
        return cls.model_validate(node)

    def render(self) -> str:
        return (
            f"[{self.repo.nameWithOwner}] {self.event_type.upper()} #{self.number}:"
            f" {self.title} (@{self.author.login})\n<{self.url}>"
        )

    @property
    def is_mine(self) -> bool:
        return self.author.isViewer

    def __str__(self) -> str:
        return f"{self.event_type} #{self.number}: {self.title!r}"


class NewIssueEvent(NewIssueoidEvent):
    CONNECTION: ClassVar[Object] = ISSUE_CONNECTION
    LAST_CONNECTION: ClassVar[Object] = ISSUE_LAST_CONNECTION
    event_type: Literal["issue"] = "issue"


class NewPREvent(NewIssueoidEvent):
    CONNECTION: ClassVar[Object] = PR_CONNECTION
    LAST_CONNECTION: ClassVar[Object] = PR_LAST_CONNECTION
    event_type: Literal["pr"] = "pr"


class NewDiscussionEvent(NewIssueoidEvent):
    CONNECTION: ClassVar[Object] = DISCUSSION_CONNECTION
    LAST_CONNECTION: ClassVar[Object] = DISCUSSION_LAST_CONNECTION
    event_type: Literal["discussion"] = "discussion"


class NewReleaseEvent(RepoActivity):
    CONNECTION: ClassVar[Object] = RELEASE_CONNECTION
    LAST_CONNECTION: ClassVar[Object] = RELEASE_LAST_CONNECTION
    event_type: Literal["release"] = "release"
    name: Optional[str] = None
    tagName: str
    author: Optional[User] = None
    description: Optional[str] = None
    descriptionHTML: Optional[str] = None
    isDraft: bool
    isPrerelease: bool
    url: str

    @field_validator("description")
    @classmethod
    def _dos2unix(cls, v: Optional[str]) -> Optional[str]:
        return dos2unix(v) if v is not None else None

    @classmethod
    def from_node(cls, repo: Repository, node: dict[str, Any]) -> NewReleaseEvent:
        log.debug("Constructing %s from node: %s", cls.__name__, json.dumps(node))
        node["timestamp"] = node.pop("createdAt")
        node["repo"] = repo
        return cls.model_validate(node)

    def render(self) -> str:
        s = f"[{self.repo.nameWithOwner}] RELEASE {self.tagName}"
        if self.isDraft:
            s += " [draft]"
        if self.isPrerelease:
            s += " [prerelease]"
        if self.name:
            s += f": {self.name}"
        if self.author is not None:
            s += f" (@{self.author.login})"
        s += f"\n<{self.url}>"
        if self.description:
            s += "\n" + reply_quote(self.description).rstrip("\n")
        return s

    @property
    def is_mine(self) -> bool:
        return self.author is not None and self.author.isViewer

    def __str__(self) -> str:
        return f"release {self.tagName}: {self.name!r}"


class NewTagEvent(RepoActivity):
    CONNECTION: ClassVar[Object] = TAG_CONNECTION
    LAST_CONNECTION: ClassVar[Object] = TAG_LAST_CONNECTION
    event_type: Literal["tag"] = "tag"
    name: str
    user: Optional[User]

    @classmethod
    def from_node(cls, repo: Repository, node: dict[str, Any]) -> NewTagEvent:
        log.debug("Constructing %s from node: %s", cls.__name__, json.dumps(node))
        target = node.pop("target")
        if target["__typename"] == "Commit":
            if target["committedDate"] is None:
                raise BogusEventError(
                    "Commit for tag {node['name']} does not have committedDate set"
                )
            node["timestamp"] = target["committedDate"]
            if target["author"] is not None:
                node["user"] = target["author"]["user"]
            else:
                node["user"] = None
        elif target["__typename"] == "Tag":
            if target["tagger"] is None or target["tagger"]["date"] is None:
                raise BogusEventError("Tag {node['name']} does not have date set")
            node["timestamp"] = target["tagger"]["date"]
            node["user"] = target["tagger"]["user"]
        else:
            raise BogusEventError(
                f"Tag {node['name']} is neither an annotated tag nor a pointer"
                " to a commit"
            )
        node["repo"] = repo
        return cls.model_validate(node)

    def render(self) -> str:
        s = f"[{self.repo.nameWithOwner}] TAG {self.name}"
        if self.user is not None:
            s += f" (@{self.user.login})"
        s += f"\n<{self.repo.url}/releases/tag/{self.name}>"
        return s

    @property
    def is_mine(self) -> bool:
        return self.user is not None and self.user.isViewer

    def __str__(self) -> str:
        return f"tag {self.name}"


class NewStarEvent(RepoActivity):
    CONNECTION: ClassVar[Object] = STAR_CONNECTION
    LAST_CONNECTION: ClassVar[Object] = STAR_LAST_CONNECTION
    event_type: Literal["star"] = "star"
    user: User

    @classmethod
    def from_node(cls, repo: Repository, node: dict[str, Any]) -> NewStarEvent:
        log.debug("Constructing %s from node: %s", cls.__name__, json.dumps(node))
        node["timestamp"] = node.pop("starredAt")
        node["repo"] = repo
        return cls.model_validate(node)

    def render(self) -> str:
        return f"★ @{self.user.login} starred {self.repo.nameWithOwner}"

    @property
    def is_mine(self) -> bool:
        return self.user.isViewer

    def __str__(self) -> str:
        return f"star by @{self.user.login}"


class NewForkEvent(RepoActivity):
    CONNECTION: ClassVar[Object] = FORK_CONNECTION
    LAST_CONNECTION: ClassVar[Object] = FORK_LAST_CONNECTION
    event_type: Literal["fork"] = "fork"
    fork: Repository

    @classmethod
    def from_node(cls, repo: Repository, node: dict[str, Any]) -> NewForkEvent:
        log.debug("Constructing %s from node: %s", cls.__name__, json.dumps(node))
        timestamp = node.pop("createdAt")
        return cls(timestamp=timestamp, repo=repo, fork=node)

    def render(self) -> str:
        return (
            f"@{self.fork.owner.login} forked {self.repo.nameWithOwner}\n"
            f"<{self.fork.url}>"
        )

    @property
    def is_mine(self) -> bool:
        return self.fork.owner.isViewer

    def __str__(self) -> str:
        return f"fork by @{self.fork.owner.login}"


class RepoTrackedEvent(Event):
    event_type: Literal["tracked"] = "tracked"

    def render(self) -> str:
        s = f"Now tracking repository {self.repo.nameWithOwner}\n<{self.repo.url}>"
        if self.repo.description:
            s += "\n" + reply_quote(self.repo.description).rstrip("\n")
        return s


class RepoUntrackedEvent(Event):
    event_type: Literal["untracked"] = "untracked"

    def render(self) -> str:
        return f"No longer tracking repository {self.repo.nameWithOwner}"


class RepoRenamedEvent(Event):
    event_type: Literal["renamed"] = "renamed"
    old_repo: Repository

    def render(self) -> str:
        return (
            f"Repository renamed: {self.old_repo.nameWithOwner} →"
            f" {self.repo.nameWithOwner}"
        )


class ActivityType(Enum):
    ISSUE = ("issue", "issues", NewIssueEvent)
    PR = ("pr", "pullRequests", NewPREvent)
    DISCUSSION = ("discussion", "discussions", NewDiscussionEvent)
    RELEASE = ("release", "releases", NewReleaseEvent)
    TAG = ("tag", "tags", NewTagEvent)
    STAR = ("star", "stargazers", NewStarEvent)
    FORK = ("fork", "forks", NewForkEvent)

    def __new__(cls, value: str, _api_name: str, _event_cls: type) -> ActivityType:
        obj = object.__new__(cls)
        obj._value_ = value
        return obj  # type: ignore[no-any-return]

    def __init__(
        self, _value: str, api_name: str, event_cls: type[RepoActivity]
    ) -> None:
        self.api_name = api_name
        self.event_cls = event_cls


CursorDict = Dict[
    Annotated[ActivityType, PlainSerializer(lambda t: t.value)], Optional[str]
]



from __future__ import annotations
import logging
from pathlib import Path
import platform
import re
from typing import TypeVar
from platformdirs import user_state_path
from pydantic import AnyHttpUrl, TypeAdapter
import requests
from . import __url__, __version__

T = TypeVar("T")


log = logging.getLogger(__package__)

MAIL_USER_AGENT = f"reponews/{__version__} ({__url__})"

HTTP_USER_AGENT = "reponews/{} ({}) requests/{} {}/{}".format(
    __version__,
    __url__,
    requests.__version__,
    platform.python_implementation(),
    platform.python_version(),
)


def mkalias(s: str) -> str:
    return s.replace("_", "-")


def get_default_state_file() -> Path:
    return user_state_path("reponews", "jwodder") / "state.json"


def dos2unix(s: str) -> str:
    return re.sub(r"\r\n?", "\n", s)


def default_api_url() -> AnyHttpUrl:
    adapter = TypeAdapter(AnyHttpUrl)
    return adapter.validate_python("https://api.github.com/graphql")


class NotFoundError(Exception):
    pass


class BogusEventError(Exception):
    pass


class UserError(Exception):
    pass
